"""
Manage input-output data from and to a target Excel file.
"""

import os
import numpy as np
import pandas as pd
from dataclasses import dataclass, field

from pyscnomics.io.config import (
    GeneralConfigData,
    FiscalConfigData,
    OilLiftingData,
    GasLiftingData,
    LPGPropaneLiftingData,
    LPGButaneLiftingData,
    SulfurLiftingData,
    ElectricityLiftingData,
    CO2LiftingData,
    CapitalCostData,
    IntangibleCostData,
    OPEXData,
    ASRCostData,
    PSCCostRecoveryData,
    PSCGrossSplitData,
    SensitivityData,
    MonteCarloData,
    OptimizationData,
)


class SpreadsheetException(Exception):
    """Exception to raise for a misuse of Spreadsheet class"""

    pass


@dataclass
class Spreadsheet:
    """
    Load and prepare data from a target Excel file.

    Attributes
    ----------
    workbook_to_read: str
        The name of the target Excel file. Must be given in '.xlsb' format.
        Defaults to None.
    directory_location: str
        The directory location of workbook.
    """

    workbook_to_read: str = field(default=None)
    directory_location: str = field(default=None)

    # Attribute associated with directory location
    load_dir: str = field(default=None, init=False)

    # Attributes associated with loading data from a target Excel file
    sheets_name: list = field(default=None, init=False, repr=False)
    sheets_loaded: list = field(default=None, init=False, repr=False)
    data_loaded: dict = field(default=None, init=False, repr=False)

    # Attributes associated with config data
    general_config_data: GeneralConfigData = field(default=None, init=False)
    fiscal_config_data: FiscalConfigData = field(default=None, init=False)

    # Attributes associated with lifting data
    oil_lifting_data: OilLiftingData = field(default=None, init=False)
    gas_lifting_data: GasLiftingData = field(default=None, init=False)
    lpg_propane_lifting_data: LPGPropaneLiftingData = field(default=None, init=False)
    lpg_butane_lifting_data: LPGButaneLiftingData = field(default=None, init=False)
    sulfur_lifting_data: SulfurLiftingData = field(default=None, init=False)
    electricity_lifting_data: ElectricityLiftingData = field(default=None, init=False)
    co2_lifting_data: CO2LiftingData = field(default=None, init=False)

    # Attributes associated with cost data
    capital_cost_data: CapitalCostData = field(default=None, init=False)
    intangible_cost_data: IntangibleCostData = field(default=None, init=False)
    opex_data: OPEXData = field(default=None, init=False)
    asr_cost_data: ASRCostData = field(default=None, init=False)

    # Attributes associated with contract data
    psc_cr_data: PSCCostRecoveryData = field(default=None, init=False)
    psc_gs_data: PSCGrossSplitData = field(default=None, init=False)
    psc_transition_cr_to_cr: dict = field(default=None, init=False)
    psc_transition_cr_to_gs: dict = field(default=None, init=False)
    psc_transition_gs_to_gs: dict = field(default=None, init=False)
    psc_transition_gs_to_cr: dict = field(default=None, init=False)

    # Fill in the attributes associated with additional functionality
    optimization_data: OptimizationData = field(default=None, init=False)
    sensitivity_data: SensitivityData = field(default=None, init=False)
    montecarlo_data: MonteCarloData = field(default=None, init=False)

    def __post_init__(self):
        # Configure attribute workbook_to_read
        if self.workbook_to_read is None:
            self.workbook_to_read = "Workbook.xlsb"

        if self.workbook_to_read is not None:
            if not isinstance(self.workbook_to_read, str):
                raise SpreadsheetException(
                    f"Excel filename must be provided in str format. "
                    f"{self.workbook_to_read} ({self.workbook_to_read.__class__.__qualname__}) "
                    f"is not a str format."
                )

        # Configure attribute directory_location
        if self.directory_location is None:
            self.load_dir = os.path.join(os.getcwd(), self.workbook_to_read)

        if self.directory_location is not None:
            self.load_dir = os.path.join(self.directory_location, self.workbook_to_read)

    def read_from_excel(self) -> None:
        """
        Reads data from a target Excel file specified by 'workbook_to_read' attribute.

        Returns
        -------
        None
            This method updates class attributes by storing information
            associated with the loaded data.

        Notes
        -----
        The core procedures are as follows:
        (1) Identify the directory location of the target Excel file,
        (2) From the target Excel file, identify the worksheets,
        (3) Load data from the target Excel file, capturing all necessary worksheets.
        """
        # Identify worksheets in the target Excel file
        excel = pd.ExcelFile(self.load_dir)
        self.sheets_name = excel.sheet_names

        # Identify the loaded sheets
        self.sheets_loaded = [
            i for i in self.sheets_name
            if i not in [
                "Cover",
                "UserGuide",
                "References",
                "ChartDATA",
                "ORETZS",
                "Summary",
                "Monte Carlo",
                "Result Table CR",
                "Result Table GS",
                "Result Table Base Project",
                "Result Table GS (2)",
                "Result Table CR (2)",
            ]
        ]

        # Load data from all worksheets
        self.data_loaded = {
            key: pd.read_excel(
                excel,
                sheet_name=key,
                skiprows=1,
                index_col=None,
                header=None,
                engine="pyxlsb",
            )
            for key in self.sheets_loaded
        }

    def _get_general_config_data(self) -> GeneralConfigData:
        """
        Extracts and prepares general configuration data from the 'General Config' sheet.

        Returns
        -------
        GeneralConfigData
            An instance of the GeneralConfigData class containing the extracted data
            stored as its attributes.

        Notes
        -----
        The procedures are as follows:
        (1) In the associated dataframe, convert NaN values to None,
        (2) Assign data to their associated attributes,
        (3) Return a new instance of GeneralConfigData with filled attributes.
        """
        # Step #1 (See 'Notes' section in the docstring)
        general_config_data_loaded = self.data_loaded["General Config"].replace(np.nan, None)

        # Step #2 - Step #3 (See 'Notes' section in the docstring)
        type_of_contract = general_config_data_loaded.iloc[0, 2]
        discount_rate_start_year = general_config_data_loaded.iloc[1, 2]
        discount_rate = general_config_data_loaded.iloc[2, 2]
        inflation_rate_applied_to = general_config_data_loaded.iloc[3, 2]
        start_date_project = general_config_data_loaded.iloc[4, 2]
        end_date_project = general_config_data_loaded.iloc[5, 2]
        start_date_project_second = general_config_data_loaded.iloc[6, 2]
        end_date_project_second = general_config_data_loaded.iloc[7, 2]
        oil_onstream_date = general_config_data_loaded.iloc[8, 2]
        gas_onstream_date = general_config_data_loaded.iloc[9, 2]
        oil_onstream_date_second = general_config_data_loaded.iloc[10, 2]
        gas_onstream_date_second = general_config_data_loaded.iloc[11, 2]
        gsa_number = general_config_data_loaded.iloc[13, 7]
        number_active_fluid = {
            "Oil": int(general_config_data_loaded.iloc[12, 3]),
            "Gas": int(general_config_data_loaded.iloc[13, 3]),
            "LPG Propane": int(general_config_data_loaded.iloc[14, 3]),
            "LPG Butane": int(general_config_data_loaded.iloc[15, 3]),
            "CO2": int(general_config_data_loaded.iloc[16, 3]),
            "Sulfur": int(general_config_data_loaded.iloc[17, 3]),
            "Electricity": int(general_config_data_loaded.iloc[18, 3]),
        }

        return GeneralConfigData(
            start_date_project=start_date_project,
            end_date_project=end_date_project,
            start_date_project_second=start_date_project_second,
            end_date_project_second=end_date_project_second,
            type_of_contract=type_of_contract,
            oil_onstream_date=oil_onstream_date,
            gas_onstream_date=gas_onstream_date,
            oil_onstream_date_second=oil_onstream_date_second,
            gas_onstream_date_second=gas_onstream_date_second,
            discount_rate_start_year=discount_rate_start_year,
            discount_rate=discount_rate,
            inflation_rate_applied_to=inflation_rate_applied_to,
            gsa_number=gsa_number,
            number_active_fluid=number_active_fluid,
        )

    def _get_fiscal_config_data(self) -> FiscalConfigData:
        """
        Extracts and prepares fiscal configuration data from the 'Fiscal Config' sheet.

        Returns
        -------
        FiscalConfigData
            An instance of the FiscalConfigData class containing the extracted data
            stored as its attributes.

        Notes
        -----
        The procedures are as follows:
        (1) Slice the data, only capture columns that contain necessary data,
        (2) Assign data to their associated attributes,
        (3) Return a new instance of FiscalConfigData with filled attributes.
        """
        # Step #1 (See 'Notes' section in the docstring)
        fiscal_config_data_loaded = self.data_loaded["Fiscal Config"].iloc[:, 1:]

        # Step #2 (See 'Notes' section in the docstring)
        # Prepare attributes associated with tax
        tax_mode = {
            "PSC 1": fiscal_config_data_loaded.iloc[0, 1],
            "PSC 2": fiscal_config_data_loaded.iloc[118, 1],
        }

        tax_rate_init = {
            "PSC 1": fiscal_config_data_loaded.iloc[1, 1],
            "PSC 2": fiscal_config_data_loaded.iloc[119, 1],
        }

        multi_tax_init = {
            "PSC 1": {
                "year": fiscal_config_data_loaded.iloc[3:10, 0].to_numpy(),
                "rate": fiscal_config_data_loaded.iloc[3:10, 1].to_numpy(),
            },
            "PSC 2": {
                "year": fiscal_config_data_loaded.iloc[122:128, 0].to_numpy(),
                "rate": fiscal_config_data_loaded.iloc[122:128, 1].to_numpy(),
            }
        }

        tax_payment_config = {
            "PSC 1": fiscal_config_data_loaded.iloc[12, 1],
            "PSC 2": fiscal_config_data_loaded.iloc[130, 1],
        }

        # Prepare attribute associated with ASR config
        asr_future_rate = {
            "PSC 1": fiscal_config_data_loaded.iloc[15, 1],
            "PSC 2": fiscal_config_data_loaded.iloc[133, 1],
        }

        # Prepare attributes associated with depreciation config
        depreciation_method = {
            "PSC 1": fiscal_config_data_loaded.iloc[18, 1],
            "PSC 2": fiscal_config_data_loaded.iloc[136, 1],
        }

        decline_factor = {
            "PSC 1": fiscal_config_data_loaded.iloc[19, 1],
            "PSC 2": fiscal_config_data_loaded.iloc[137, 1],
        }

        # Prepare attributes associated with inflation config
        inflation_rate_mode = {
            "PSC 1": fiscal_config_data_loaded.iloc[22, 1],
            "PSC 2": fiscal_config_data_loaded.iloc[140, 1],
        }

        inflation_rate_init = {
            "PSC 1": fiscal_config_data_loaded.iloc[23, 1],
            "PSC 2": fiscal_config_data_loaded.iloc[141, 1],
        }

        multi_inflation_init = {
            "PSC 1": {
                "year": fiscal_config_data_loaded.iloc[26:46, 0].to_numpy(),
                "rate": fiscal_config_data_loaded.iloc[26:46, 1].to_numpy(),
            },
            "PSC 2": {
                "year": fiscal_config_data_loaded.iloc[144:164, 0].to_numpy(),
                "rate": fiscal_config_data_loaded.iloc[144:164, 1].to_numpy(),
            }
        }

        # Prepare attributes associated with VAT config
        vat_mode = {
            "PSC 1": fiscal_config_data_loaded.iloc[48, 1],
            "PSC 2": fiscal_config_data_loaded.iloc[166, 1],
        }

        vat_rate_init = {
            "PSC 1": fiscal_config_data_loaded.iloc[49, 1],
            "PSC 2": fiscal_config_data_loaded.iloc[167, 1],
        }

        multi_vat_init = {
            "PSC 1": {
                "year": fiscal_config_data_loaded.iloc[52:72, 0].to_numpy(),
                "rate": fiscal_config_data_loaded.iloc[52:72, 1].to_numpy(),
            },
            "PSC 2": {
                "year": fiscal_config_data_loaded.iloc[170:190, 0].to_numpy(),
                "rate": fiscal_config_data_loaded.iloc[170:190, 1].to_numpy(),
            }
        }

        # Prepare attributes associated with LBT config
        lbt_mode = {
            "PSC 1": fiscal_config_data_loaded.iloc[74, 1],
            "PSC 2": fiscal_config_data_loaded.iloc[192, 1],
        }

        lbt_rate_init = {
            "PSC 1": fiscal_config_data_loaded.iloc[75, 1],
            "PSC 2": fiscal_config_data_loaded.iloc[193, 1],
        }

        multi_lbt_init = {
            "PSC 1": {
                "year": fiscal_config_data_loaded.iloc[78:98, 0].to_numpy(),
                "rate": fiscal_config_data_loaded.iloc[78:98, 1].to_numpy(),
            },
            "PSC 2": {
                "year": fiscal_config_data_loaded.iloc[196:216, 0].to_numpy(),
                "rate": fiscal_config_data_loaded.iloc[196:216, 1].to_numpy(),
            }
        }

        # Prepare attributes associated with discount config
        vat_discount = {
            "PSC 1": fiscal_config_data_loaded.iloc[100, 1],
            "PSC 2": fiscal_config_data_loaded.iloc[218, 1],
        }

        lbt_discount = {
            "PSC 1": fiscal_config_data_loaded.iloc[101, 1],
            "PSC 2": fiscal_config_data_loaded.iloc[219, 1],
        }

        # Prepare attributes associated with NPV config
        npv_mode = fiscal_config_data_loaded.iloc[104, 1]
        discounting_mode = fiscal_config_data_loaded.iloc[105, 1]

        # Prepare attributes associated with other revenue config
        sulfur_revenue_config = {
            "PSC 1": fiscal_config_data_loaded.iloc[108, 1],
            "PSC 2": fiscal_config_data_loaded.iloc[222, 1],
        }

        electricity_revenue_config = {
            "PSC 1": fiscal_config_data_loaded.iloc[109, 1],
            "PSC 2": fiscal_config_data_loaded.iloc[223, 1],
        }

        co2_revenue_config = {
            "PSC 1": fiscal_config_data_loaded.iloc[110, 1],
            "PSC 2": fiscal_config_data_loaded.iloc[224, 1],
        }

        # Prepare attribute associated with transition config
        transferred_unrec_cost = fiscal_config_data_loaded.iloc[115, 1]

        # Prepare attribute associated with sunk cost config
        sunk_cost_reference_year = fiscal_config_data_loaded.iloc[230, 1]

        # Step #3 (See 'Notes' section in the docstring)
        return FiscalConfigData(
            tax_mode=tax_mode,
            tax_rate_init=tax_rate_init,
            multi_tax_init=multi_tax_init,
            tax_payment_config=tax_payment_config,
            asr_future_rate=asr_future_rate,
            depreciation_method=depreciation_method,
            decline_factor=decline_factor,
            inflation_rate_mode=inflation_rate_mode,
            inflation_rate_init=inflation_rate_init,
            multi_inflation_init=multi_inflation_init,
            vat_mode=vat_mode,
            vat_rate_init=vat_rate_init,
            multi_vat_init=multi_vat_init,
            lbt_mode=lbt_mode,
            lbt_rate_init=lbt_rate_init,
            multi_lbt_init=multi_lbt_init,
            vat_discount=vat_discount,
            lbt_discount=lbt_discount,
            npv_mode=npv_mode,
            discounting_mode=discounting_mode,
            sulfur_revenue_config=sulfur_revenue_config,
            electricity_revenue_config=electricity_revenue_config,
            co2_revenue_config=co2_revenue_config,
            transferred_unrec_cost=transferred_unrec_cost,
            sunk_cost_reference_year=sunk_cost_reference_year,
            project_years=self.general_config_data.project_years,
            type_of_contract=self.general_config_data.type_of_contract,
        )

    def _get_oil_lifting_data(self) -> OilLiftingData:
        """
        Retrieves oil lifting data based on available sheets.

        Returns
        -------
        OilLiftingData
            An instance of the OilLiftingData class.

        Notes
        -----
        The undertaken procedures are as follows:
        (1) Filter attribute self.sheets_loaded for sheets that contain 'Prod Oil' data,
            then assigned it as local variable named 'oil_data_available',
        (2) Load the data associated with oil_data, then store it in the variable
            named 'oil_data_loaded_init',
        (3) Undertake data cleansing: remove all rows which column 'prod_year' is NaN.
            Store the results in the variable named 'oil_data_loaded',
        (4) Create a dictionary named 'oil_data' to store the necessary data from 'oil_data_loaded',
        (5) Return an instance of OilLiftingData to store the oil data appropriately
            as its attributes.
        """
        # Step #1 (See 'Notes' section in the docstring)
        oil_data_available = list(filter(lambda i: "Prod Oil" in i, self.sheets_loaded))

        # Step #2 (See 'Notes' section in the docstring)
        oil_data_loaded_init = {ws: self.data_loaded[ws] for ws in oil_data_available}

        # Step #3 (See 'Notes' section in the docstring)
        oil_data_loaded = {
            ws: oil_data_loaded_init[ws] if oil_data_loaded_init[ws].empty
            else oil_data_loaded_init[ws][~pd.isna(oil_data_loaded_init[ws].iloc[:, 0])].copy()
            for ws in oil_data_available
        }

        # Step #4 (See 'Notes' section in the docstring)
        oil_attrs = [
            "prod_year",
            "oil_lifting_rate",
            "oil_price",
            "condensate_lifting_rate",
            "condensate_price",
        ]

        oil_data = {
            key: {
                ws: None if oil_data_loaded[ws].empty
                else oil_data_loaded[ws].iloc[:, i].to_numpy()
                for ws in oil_data_available
            }
            for i, key in enumerate(oil_attrs)
        }

        # Step #5 (See 'Notes' section in the docstring)
        return OilLiftingData(
            prod_year_init=oil_data["prod_year"],
            oil_lifting_rate_init=oil_data["oil_lifting_rate"],
            oil_price_init=oil_data["oil_price"],
            condensate_lifting_rate_init=oil_data["condensate_lifting_rate"],
            condensate_price_init=oil_data["condensate_price"],
            active_oil=self.general_config_data.number_active_fluid["Oil"],
            project_duration=self.general_config_data.project_duration,
            project_years=self.general_config_data.project_years,
            type_of_contract=self.general_config_data.type_of_contract,
            end_date_project=self.general_config_data.end_date_project,
            start_date_project_second=self.general_config_data.start_date_project_second,
        )

    def _get_gas_lifting_data(self) -> GasLiftingData:
        """
        Retrieves gas lifting data based on available sheets.

        Returns
        -------
        GasLiftingData
            An instance of GasLiftingData class.

        Notes
        -----
        The undertaken procedures are as follows:
        (1) Filter attribute self.sheets_loaded for sheets that contain 'Prod Gas' data,
            then assigned it as local variable named 'gas_data_available',
        (2) Load the data associated with gas, then store it in the variable
            named 'gas_data_loaded_init',
        (3) Undertake data cleansing: remove all rows which column 'prod_year' is NaN.
            Store the results in the variable named 'gas_data_loaded',
        (4) Create a dictionary named 'gas_data_general' to store information related to
            'prod_year' and 'prod_rate' from 'gas_data_loaded',
        (5) -   Define variable named 'gsa_id'. This variable stores information regarding
                the associated column indices in 'gas_data_loaded' that corresponds to
                'lifting_rate', 'ghv', and 'price'. Variable 'gsa_id' is a 2D array: the
                zeroth axis depicts indices of 'lifting_rate', 'ghv', and 'price' while the
                first axis manifest the number of available GSA,
            -   Create a dictionary named 'gas_data_gsa' to store information related to
                'lifting_rate', 'ghv', and 'price',
        (6) Return an instance of GasLiftingData to store the gas data appropriately
            as its attributes.
        """
        # Step #1 (See 'Notes' section in the docstring)
        gas_data_available = list(filter(lambda i: "Prod Gas" in i, self.sheets_loaded))

        # Step #2 (See 'Notes' section in the docstring)
        gas_data_loaded_init = {ws: self.data_loaded[ws] for ws in gas_data_available}

        # Step #3 (See 'Notes' section in the docstring)
        gas_data_loaded = {
            ws: gas_data_loaded_init[ws] if gas_data_loaded_init[ws].empty
            else (
                gas_data_loaded_init[ws]
                [~pd.isna(gas_data_loaded_init[ws].iloc[:, 0])]
                .copy()
            )
            for ws in gas_data_available
        }

        # Step #4 (See 'Notes' section in the docstring)
        gas_attrs_general = ["prod_year", "prod_rate"]
        gas_data_general = {
            key: {
                ws: None if gas_data_loaded[ws].empty
                else gas_data_loaded[ws].iloc[:, i].to_numpy()
                for ws in gas_data_available
            }
            for i, key in enumerate(gas_attrs_general)
        }

        # Step #5 (See 'Notes' section in the docstring)
        gsa_number = self.general_config_data.gsa_number
        gsa_variables = ["GSA {0}".format(i + 1) for i in range(gsa_number)]
        gas_attrs_gsa = ["lifting_rate", "ghv", "price"]

        gsa_id = (
            np.repeat(np.arange(len(gas_attrs_gsa))[:, np.newaxis] + 2, gsa_number, axis=1)
            + (np.arange(gsa_number) * len(gas_attrs_gsa))[np.newaxis, :]
        )

        gas_data_gsa = {
            key: {
                ws: {
                    gsa: None if gas_data_loaded[ws].empty
                    else gas_data_loaded[ws].iloc[:, gsa_id[i, j]].to_numpy()
                    for gsa, j in zip(gsa_variables, range(gsa_number))
                }
                for ws in gas_data_available
            }
            for i, key in enumerate(gas_attrs_gsa)
        }

        # Step #6 (See 'Notes' section in the docstring)
        return GasLiftingData(
            gsa_number=gsa_number,
            active_gas=self.general_config_data.number_active_fluid["Gas"],
            prod_year_init=gas_data_general["prod_year"],
            prod_rate_init=gas_data_general["prod_rate"],
            lifting_rate_init=gas_data_gsa["lifting_rate"],
            ghv_init=gas_data_gsa["ghv"],
            price_init=gas_data_gsa["price"],
            project_duration=self.general_config_data.project_duration,
            project_years=self.general_config_data.project_years,
            type_of_contract=self.general_config_data.type_of_contract,
            end_date_project=self.general_config_data.end_date_project,
            start_date_project_second=self.general_config_data.start_date_project_second,
        )

    def _get_lpg_propane_lifting_data(self) -> LPGPropaneLiftingData:
        """
        Retrieves LPG propane lifting data based on available sheets.

        Returns
        -------
        LPGPropaneLiftingData
            An instance of the LPGPropaneLiftingData class.

        Notes
        -----
        The undertaken procedures are as follows:
        (1) Filter attribute self.sheets_loaded for sheets that contain 'Prod LPG Propane'
            data, then assigned it as local variable named 'lpg_propane_data_available',
        (2) Load the data associated with lpg_propane_data, then store it in the variable
            named 'lpg_propane_data_loaded_init',
        (3) Undertake data cleansing: remove all rows which column 'prod_year' is NaN.
            Store the results in the variable named 'lpg_propane_data_loaded',
        (4) Create a dictionary named 'lpg_propane_data' to store the necessary data from
            'lpg_propane_data_loaded',
        (5) Return an instance of LPGPropaneLiftingData to store the lpg propane data
            appropriately as its attributes.
        """
        # Step #1 (See 'Notes' section in the docstring)
        lpg_propane_data_available = list(
            filter(lambda i: "Prod LPG Propane" in i, self.sheets_loaded)
        )

        # Step #2 (See 'Notes' section in the docstring)
        lpg_propane_data_loaded_init = {
            ws: self.data_loaded[ws] for ws in lpg_propane_data_available
        }

        # Step #3 (See 'Notes' section in the docstring)
        lpg_propane_data_loaded = {
            ws: lpg_propane_data_loaded_init[ws] if lpg_propane_data_loaded_init[ws].empty
            else (
                lpg_propane_data_loaded_init[ws]
                [~pd.isna(lpg_propane_data_loaded_init[ws].iloc[:, 0])]
                .copy()
            )
            for ws in lpg_propane_data_available
        }

        # Step #4 (See 'Notes' section in the docstring)
        lpg_propane_attrs = ["prod_year", "lifting_rate", "price"]

        lpg_propane_data = {
            key: {
                ws: None if lpg_propane_data_loaded[ws].empty
                else lpg_propane_data_loaded[ws].iloc[:, i].to_numpy()
                for ws in lpg_propane_data_available
            }
            for i, key in enumerate(lpg_propane_attrs)
        }

        # Step #5 (See 'Notes' section in the docstring)
        return LPGPropaneLiftingData(
            prod_year_init=lpg_propane_data["prod_year"],
            lifting_rate_init=lpg_propane_data["lifting_rate"],
            price_init=lpg_propane_data["price"],
            active_lpgpropane=self.general_config_data.number_active_fluid["LPG Propane"],
            project_duration=self.general_config_data.project_duration,
            project_years=self.general_config_data.project_years,
            type_of_contract=self.general_config_data.type_of_contract,
            end_date_project=self.general_config_data.end_date_project,
            start_date_project_second=self.general_config_data.start_date_project_second,
        )

    def _get_lpg_butane_lifting_data(self) -> LPGButaneLiftingData:
        """
        Retrieves LPG butane lifting data based on available sheets.

        Returns
        -------
        LPGButaneLiftingData
            An instance of the LPGButaneLiftingData class.

        Notes
        -----
        The undertaken procedures are as follows:
        (1) Filter attribute self.sheets_loaded for sheets that contain 'Prod LPG Butane'
            data, then assigned it as local variable named 'lpg_butane_data_available',
        (2) Load the data associated with lpg_butane_data, then store it in the variable
            named 'lpg_butane_data_loaded_init',
        (3) Undertake data cleansing: remove all rows which column 'prod_year' is NaN.
            Store the results in the variable named 'lpg_butane_data_loaded',
        (4) Create a dictionary named 'lpg_butane_data' to store the necessary data from
            'lpg_butane_data_loaded',
        (5) Return an instance of LPGButaneLiftingData to store the lpg butane data
            appropriately as its attributes.
        """
        # Step #1 (See 'Notes' section in the docstring)
        lpg_butane_data_available = list(
            filter(lambda i: "Prod LPG Butane" in i, self.sheets_loaded)
        )

        # Step #2 (See 'Notes' section in the docstring)
        lpg_butane_data_loaded_init = {
            ws: self.data_loaded[ws] for ws in lpg_butane_data_available
        }

        # Step #3 (See 'Notes' section in the docstring)
        lpg_butane_data_loaded = {
            ws: lpg_butane_data_loaded_init[ws] if lpg_butane_data_loaded_init[ws].empty
            else (
                lpg_butane_data_loaded_init[ws]
                [~pd.isna(lpg_butane_data_loaded_init[ws].iloc[:, 0])]
                .copy()
            )
            for ws in lpg_butane_data_available
        }

        # Step #4 (See 'Notes' section in the docstring)
        lpg_butane_attrs = ["prod_year", "lifting_rate", "price"]

        lpg_butane_data = {
            key: {
                ws: None if lpg_butane_data_loaded[ws].empty
                else lpg_butane_data_loaded[ws].iloc[:, i].to_numpy()
                for ws in lpg_butane_data_available
            }
            for i, key in enumerate(lpg_butane_attrs)
        }

        # Step #5 (See 'Notes' section in the docstring)
        return LPGButaneLiftingData(
            prod_year_init=lpg_butane_data["prod_year"],
            lifting_rate_init=lpg_butane_data["lifting_rate"],
            price_init=lpg_butane_data["price"],
            active_lpgbutane=self.general_config_data.number_active_fluid["LPG Butane"],
            project_duration=self.general_config_data.project_duration,
            project_years=self.general_config_data.project_years,
            type_of_contract=self.general_config_data.type_of_contract,
            end_date_project=self.general_config_data.end_date_project,
            start_date_project_second=self.general_config_data.start_date_project_second,
        )

    def _get_sulfur_lifting_data(self) -> SulfurLiftingData:
        """
        Retrieves sulfur lifting data based on available sheets.

        Returns
        -------
        SulfurLiftingData
            An instance of the SulfurLiftingData class.

        Notes
        -----
        The undertaken procedures are as follows:
        (1) Filter attribute self.sheets_loaded for sheets that contain 'Prod Sulfur'
            data, then assigned it as local variable named 'sulfur_data_available',
        (2) Load the data associated with sulfur data, then store it in the variable
            named 'sulfur_data_loaded_init',
        (3) Undertake data cleansing: remove all rows which column 'prod_year' is NaN.
            Store the results in the variable named 'sulfur_data_loaded',
        (4) Create a dictionary named 'sulfur_data' to store the necessary data from
            'sulfur_data_loaded',
        (5) Return an instance of SulfurLiftingData to store the sulfur data
            appropriately as its attributes.
        """
        # Step #1 (See 'Notes' section in the docstring)
        sulfur_data_available = list(filter(lambda i: "Prod Sulfur" in i, self.sheets_loaded))

        # Step #2 (See 'Notes' section in the docstring)
        sulfur_data_loaded_init = {ws: self.data_loaded[ws] for ws in sulfur_data_available}

        # Step #3 (See 'Notes' section in the docstring)
        sulfur_data_loaded = {
            ws: sulfur_data_loaded_init[ws] if sulfur_data_loaded_init[ws].empty
            else (
                sulfur_data_loaded_init[ws]
                [~pd.isna(sulfur_data_loaded_init[ws].iloc[:, 0])]
                .copy()
            )
            for ws in sulfur_data_available
        }

        # Step #4 (See 'Notes' section in the docstring)
        sulfur_data_attrs = ["prod_year", "lifting_rate", "price"]

        sulfur_data = {
            key: {
                ws: None if sulfur_data_loaded[ws].empty
                else sulfur_data_loaded[ws].iloc[:, i].to_numpy()
                for ws in sulfur_data_available
            }
            for i, key in enumerate(sulfur_data_attrs)
        }

        # Step #5 (See 'Notes' section in the docstring)
        return SulfurLiftingData(
            prod_year_init=sulfur_data["prod_year"],
            lifting_rate_init=sulfur_data["lifting_rate"],
            price_init=sulfur_data["price"],
            active_sulfur=self.general_config_data.number_active_fluid["Sulfur"],
            project_duration=self.general_config_data.project_duration,
            project_years=self.general_config_data.project_years,
            type_of_contract=self.general_config_data.type_of_contract,
            end_date_project=self.general_config_data.end_date_project,
            start_date_project_second=self.general_config_data.start_date_project_second,
        )

    def _get_electricity_lifting_data(self) -> ElectricityLiftingData:
        """
        Retrieves electricity lifting data based on available sheets.

        Returns
        -------
        ElectricityLiftingData
            An instance of the ElectricityLiftingData class.

        Notes
        -----
        The undertaken procedures are as follows:
        (1) Filter attribute self.sheets_loaded for sheets that contain 'Prod Electricity'
            data, then assigned it as local variable named 'electricity_data_available',
        (2) Load the data associated with electricity data, then store it in the variable
            named 'electricity_data_loaded_init',
        (3) Undertake data cleansing: remove all rows which column 'prod_year' is NaN.
            Store the results in the variable named 'electricity_data_loaded',
        (4) Create a dictionary named 'electricity_data' to store the necessary data from
            'electricity_data_loaded',
        (5) Return an instance of ElectricityLiftingData to store the electricity data
            appropriately as its attributes.
        """
        # Step #1 (See 'Notes' section in the docstring)
        electricity_data_available = list(filter(lambda i: "Prod Electricity" in i, self.sheets_loaded))

        # Step #2 (See 'Notes' section in the docstring)
        electricity_data_loaded_init = {
            ws: self.data_loaded[ws] for ws in electricity_data_available
        }

        # Step #3 (See 'Notes' section in the docstring)
        electricity_data_loaded = {
            ws: electricity_data_loaded_init[ws] if electricity_data_loaded_init[ws].empty
            else (
                electricity_data_loaded_init[ws]
                [~pd.isna(electricity_data_loaded_init[ws].iloc[:, 0])]
                .copy()
            )
            for ws in electricity_data_available
        }

        # Step #4 (See 'Notes' section in the docstring)
        electricity_data_attrs = ["prod_year", "lifting_rate", "price"]

        electricity_data = {
            key: {
                ws: None if electricity_data_loaded[ws].empty
                else electricity_data_loaded[ws].iloc[:, i].to_numpy()
                for ws in electricity_data_available
            }
            for i, key in enumerate(electricity_data_attrs)
        }

        # Step #5 (See 'Notes' section in the docstring)
        return ElectricityLiftingData(
            prod_year_init=electricity_data["prod_year"],
            lifting_rate_init=electricity_data["lifting_rate"],
            price_init=electricity_data["price"],
            active_electricity=self.general_config_data.number_active_fluid["Electricity"],
            project_duration=self.general_config_data.project_duration,
            project_years=self.general_config_data.project_years,
            type_of_contract=self.general_config_data.type_of_contract,
            end_date_project=self.general_config_data.end_date_project,
            start_date_project_second=self.general_config_data.start_date_project_second,
        )

    def _get_co2_lifting_data(self) -> CO2LiftingData:
        """
        Retrieves CO2 lifting data based on available sheets.

        Returns
        -------
        CO2LiftingData
            An instance of the CO2LiftingData class.

        Notes
        -----
        The undertaken procedures are as follows:
        (1) Filter attribute self.sheets_loaded for sheets that contain 'Prod CO2'
            data, then assigned it as local variable named 'co2_data_available',
        (2) Load the data associated with co2 data, then store it in the variable
            named 'co2_data_loaded_init',
        (3) Undertake data cleansing: remove all rows which column 'prod_year' is NaN.
            Store the results in the variable named 'co2_data_loaded',
        (4) Create a dictionary named 'co2_data' to store the necessary data from
            'co2_data_loaded',
        (5) Return an instance of CO2LiftingData to store the co2 data
            appropriately as its attributes.
        """
        # Step #1 (See 'Notes' section in the docstring)
        co2_data_available = list(filter(lambda i: "Prod CO2" in i, self.sheets_loaded))

        # Step #2 (See 'Notes' section in the docstring)
        co2_data_loaded_init = {ws: self.data_loaded[ws] for ws in co2_data_available}

        # Step #3 (See 'Notes' section in the docstring)
        co2_data_loaded = {
            ws: co2_data_loaded_init[ws] if co2_data_loaded_init[ws].empty
            else (
                co2_data_loaded_init[ws]
                [~pd.isna(co2_data_loaded_init[ws].iloc[:, 0])]
                .copy()
            )
            for ws in co2_data_available
        }

        # Step #4 (See 'Notes' section in the docstring)
        co2_data_attrs = ["prod_year", "lifting_rate", "price"]

        co2_data = {
            key: {
                ws: None if co2_data_loaded[ws].empty
                else co2_data_loaded[ws].iloc[:, i].to_numpy()
                for ws in co2_data_available
            }
            for i, key in enumerate(co2_data_attrs)
        }

        # Step #5 (See 'Notes' section in the docstring)
        return CO2LiftingData(
            prod_year_init=co2_data["prod_year"],
            lifting_rate_init=co2_data["lifting_rate"],
            price_init=co2_data["price"],
            active_co2=self.general_config_data.number_active_fluid["CO2"],
            project_duration=self.general_config_data.project_duration,
            project_years=self.general_config_data.project_years,
            type_of_contract=self.general_config_data.type_of_contract,
            end_date_project=self.general_config_data.end_date_project,
            start_date_project_second=self.general_config_data.start_date_project_second,
        )

    def _get_capital_cost_data(self) -> CapitalCostData:
        """
        Retrieves capital cost data based on available sheets.

        Returns
        -------
        CapitalCostData
            An instance of CapitalCostData class.

        Notes
        -----
        The core operations are as follows:
        (1) Capture capital cost data from attribute self.data_loaded, then perform the
            necessary adjustment,
        (2) Undertake data cleansing: remove all rows which column 'expense_year' is NaN.
            Store the results in a variable named 'capital_data_loaded',
        (3) Create a dictionary named 'capital_data' to store the necessary data from
            'capital_data_loaded',
        (4) Return an instance of CapitalCostData to store the capital data appropriately
            as its attributes.
        """
        # Step #1 (See 'Notes' section in the docstring)
        capital_data_loaded_init = self.data_loaded["Cost Tangible"].dropna(axis=0, how="all")

        # Step #2 (See 'Notes' section in the docstring)
        capital_data_loaded = (
            capital_data_loaded_init if capital_data_loaded_init.empty
            else capital_data_loaded_init[~pd.isna(capital_data_loaded_init.iloc[:, 0])].copy()
        )

        # Step #3 (See 'Notes' section in the docstring)
        capital_data_attrs = [
            "expense_year",
            "cost_allocation",
            "cost",
            "pis_year",
            "useful_life",
            "depreciation_factor",
            "salvage_value",
            "is_ic_applied",
            "vat_portion",
            "lbt_portion",
            "description",
        ]

        capital_data = {
            key: None if capital_data_loaded.empty
            else capital_data_loaded.iloc[:, i].to_numpy()
            for i, key in enumerate(capital_data_attrs)
        }

        # Step #4 (See 'Notes' section in the docstring)
        return CapitalCostData(
            expense_year_init=capital_data["expense_year"],
            cost_allocation=capital_data["cost_allocation"],
            cost=capital_data["cost"],
            pis_year=capital_data["pis_year"],
            useful_life=capital_data["useful_life"],
            depreciation_factor=capital_data["depreciation_factor"],
            salvage_value=capital_data["salvage_value"],
            is_ic_applied=capital_data["is_ic_applied"],
            vat_portion=capital_data["vat_portion"],
            lbt_portion=capital_data["lbt_portion"],
            description=capital_data["description"],
            project_years=self.general_config_data.project_years,
            type_of_contract=self.general_config_data.type_of_contract,
            end_date_project=self.general_config_data.end_date_project,
            start_date_project_second=self.general_config_data.start_date_project_second,
        )

    def _get_intangible_cost_data(self) -> IntangibleCostData:
        """
        Retrieves intangible cost data based on available sheets.

        Returns
        -------
        IntangibleCostData
            An instance of IntangibleCostData class.

        Notes
        -----
        The core operations are as follows:
        (1) Capture intangible cost data from attribute self.data_loaded, then perform the
            necessary adjustment,
        (2) Undertake data cleansing: remove all rows which column 'expense_year' is NaN.
            Store the results in a variable named 'intangible_data_loaded',
        (3) Create a dictionary named 'intangible_data' to store the necessary data from
            'intangible_data_loaded',
        (4) Return an instance of IntangibleCostData to store the intangible data appropriately
            as its attributes.
        """
        # Step #1 (See 'Notes' section in the docstring)
        intangible_data_loaded_init = self.data_loaded["Cost Intangible"].dropna(axis=0, how="all")

        # Step #2 (See 'Notes' section in the docstring)
        intangible_data_loaded = (
            intangible_data_loaded_init if intangible_data_loaded_init.empty
            else intangible_data_loaded_init[~pd.isna(intangible_data_loaded_init.iloc[:, 0])].copy()
        )

        # Step #3 (See 'Notes' section in the docstring)
        intangible_data_attrs = [
            "expense_year",
            "cost_allocation",
            "cost",
            "vat_portion",
            "lbt_portion",
            "description",
        ]

        intangible_data = {
            key: None if intangible_data_loaded.empty
            else intangible_data_loaded.iloc[:, i].to_numpy()
            for i, key in enumerate(intangible_data_attrs)
        }

        # Step #4 (See 'Notes' section in the docstring)
        return IntangibleCostData(
            expense_year_init=intangible_data["expense_year"],
            cost_allocation=intangible_data["cost_allocation"],
            cost=intangible_data["cost"],
            vat_portion=intangible_data["vat_portion"],
            lbt_portion=intangible_data["lbt_portion"],
            description=intangible_data["description"],
            project_years=self.general_config_data.project_years,
            type_of_contract=self.general_config_data.type_of_contract,
            end_date_project=self.general_config_data.end_date_project,
            start_date_project_second=self.general_config_data.start_date_project_second,
        )

    def _get_opex_data(self) -> OPEXData:
        """
        Retrieves opex data based on available sheets.

        Returns
        -------
        OPEXData
            An instance of OPEXData class.

        Notes
        -----
        The core operations are as follows:
        (1) Capture opex data from attribute self.data_loaded, then perform the
            necessary adjustment,
        (2) Undertake data cleansing: remove all rows which column 'expense_year' is NaN.
            Store the results in a variable named 'opex_data_loaded',
        (3) Create a dictionary named 'opex_data' to store the necessary data from
            'opex_data_loaded',
        (4) Return an instance of OPEXData to store the asr data appropriately
            as its attributes.
        """
        # Step #1 (See 'Notes' section in the docstring)
        opex_data_loaded_init = self.data_loaded["Cost OPEX"].dropna(axis=0, how="all")

        # Step #2 (See 'Notes' section in the docstring)
        opex_data_loaded = (
            opex_data_loaded_init if opex_data_loaded_init.empty
            else opex_data_loaded_init[~pd.isna(opex_data_loaded_init.iloc[:, 0])].copy()
        )

        # Step #3 (See 'Notes' section in the docstring)
        opex_data_attrs = [
            "expense_year",
            "cost_allocation",
            "fixed_cost",
            "prod_rate",
            "cost_per_volume",
            "vat_portion",
            "lbt_portion",
            "description",
        ]

        opex_data = {
            key: None if opex_data_loaded.empty
            else opex_data_loaded.iloc[:, i].to_numpy()
            for i, key in enumerate(opex_data_attrs)
        }

        # Step #4 (See 'Notes' section in the docstring)
        return OPEXData(
            expense_year_init=opex_data["expense_year"],
            cost_allocation=opex_data["cost_allocation"],
            fixed_cost=opex_data["fixed_cost"],
            prod_rate=opex_data["prod_rate"],
            cost_per_volume=opex_data["cost_per_volume"],
            vat_portion=opex_data["vat_portion"],
            lbt_portion=opex_data["lbt_portion"],
            description=opex_data["description"],
            project_years=self.general_config_data.project_years,
            type_of_contract=self.general_config_data.type_of_contract,
            end_date_project=self.general_config_data.end_date_project,
            start_date_project_second=self.general_config_data.start_date_project_second,
        )

    def _get_asr_cost_data(self) -> ASRCostData:
        """
        Retrieves ASR cost data based on available sheets.

        Returns
        -------
        ASRCostData
            An instance of ASRCostData class.

        Notes
        -----
        The core operations are as follows:
        (1) Capture asr cost data from attribute self.data_loaded, then perform the
            necessary adjustment,
        (2) Undertake data cleansing: remove all rows which column 'expense_year' is NaN.
            Store the results in a variable named 'asr_data_loaded',
        (3) Create a dictionary named 'asr_data' to store the necessary data from
            'asr_data_loaded',
        (4) Return an instance of ASRCostData to store the asr data appropriately
            as its attributes.
        """
        # Step #1 (See 'Notes' section in the docstring)
        asr_data_loaded_init = self.data_loaded["Cost ASR"].dropna(axis=0, how="all")

        # Step #2 (See 'Notes' section in the docstring)
        asr_data_loaded = (
            asr_data_loaded_init if asr_data_loaded_init.empty
            else asr_data_loaded_init[~pd.isna(asr_data_loaded_init.iloc[:, 0])].copy()
        )

        # Step #3 (See 'Notes' section in the docstring)
        asr_data_attrs = [
            "expense_year",
            "cost_allocation",
            "cost",
            "vat_portion",
            "lbt_portion",
            "description",
        ]

        asr_data = {
            key: None if asr_data_loaded.empty
            else asr_data_loaded.iloc[:, i].to_numpy()
            for i, key in enumerate(asr_data_attrs)
        }

        # Step #4 (See 'Notes' section in the docstring)
        return ASRCostData(
            expense_year_init=asr_data["expense_year"],
            cost_allocation=asr_data["cost_allocation"],
            cost=asr_data["cost"],
            vat_portion=asr_data["vat_portion"],
            lbt_portion=asr_data["lbt_portion"],
            description=asr_data["description"],
            project_years=self.general_config_data.project_years,
            type_of_contract=self.general_config_data.type_of_contract,
            end_date_project=self.general_config_data.end_date_project,
            start_date_project_second=self.general_config_data.start_date_project_second,
        )

    def _get_psc_cr_data(self) -> PSCCostRecoveryData:
        """
        Retrieves Production Sharing Contract (PSC) Cost Recovery data
        based on available sheets.

        Returns
        -------
        PSCCostRecoveryData
            An instance of PSCCostRecoveryData representing the extracted data
            stored as its attributes.

        Notes
        -----
        The procedures are as follows:
        (1) Slice the data, only capture the columns that contain necessary data,
        (2) Convert the remaining NaN values to None,
        (3) Assign data to their associated attributes,
        (4) Return a new instance of PSCCostRecoveryData with filled attributes.
        """
        # Step #1 - Step #2 (See 'Notes' section in the docstring)
        psc_cr_data_loaded = self.data_loaded["Cost Recovery Config"].replace(np.nan, None)

        # Step #3 - Step #4 (See 'Notes' section in the docstring)
        return PSCCostRecoveryData(
            oil_ftp_availability=psc_cr_data_loaded.iloc[0, 2],
            oil_ftp_is_shared=psc_cr_data_loaded.iloc[1, 2],
            oil_ftp_portion=psc_cr_data_loaded.iloc[2, 2],
            gas_ftp_availability=psc_cr_data_loaded.iloc[3, 2],
            gas_ftp_is_shared=psc_cr_data_loaded.iloc[4, 2],
            gas_ftp_portion=psc_cr_data_loaded.iloc[5, 2],
            split_type=psc_cr_data_loaded.iloc[8, 2],
            oil_ctr_pretax=psc_cr_data_loaded.iloc[9, 2],
            gas_ctr_pretax=psc_cr_data_loaded.iloc[10, 2],
            ic_availability=psc_cr_data_loaded.iloc[13, 2],
            ic_oil=psc_cr_data_loaded.iloc[14, 2],
            ic_gas=psc_cr_data_loaded.iloc[15, 2],
            oil_cr_cap_rate=psc_cr_data_loaded.iloc[18, 2],
            gas_cr_cap_rate=psc_cr_data_loaded.iloc[19, 2],
            dmo_is_weighted=psc_cr_data_loaded.iloc[22, 2],
            oil_dmo_holiday=psc_cr_data_loaded.iloc[25, 2],
            oil_dmo_period=psc_cr_data_loaded.iloc[26, 2],
            oil_dmo_start_production=psc_cr_data_loaded.iloc[27, 2],
            oil_dmo_volume=psc_cr_data_loaded.iloc[28, 2],
            oil_dmo_fee=psc_cr_data_loaded.iloc[29, 2],
            gas_dmo_holiday=psc_cr_data_loaded.iloc[32, 2],
            gas_dmo_period=psc_cr_data_loaded.iloc[33, 2],
            gas_dmo_start_production=psc_cr_data_loaded.iloc[34, 2],
            gas_dmo_volume=psc_cr_data_loaded.iloc[35, 2],
            gas_dmo_fee=psc_cr_data_loaded.iloc[36, 2],
            rc_split_init=psc_cr_data_loaded.iloc[41:48, 1:].to_numpy(),
            icp_sliding_scale_init=psc_cr_data_loaded.iloc[52:60, 1:].to_numpy(),
            indicator_rc_split_sliding_scale_init=psc_cr_data_loaded.iloc[63:83, 1:3].to_numpy(),
        )

    def _get_psc_gs_data(self) -> PSCGrossSplitData:
        """
        Retrieves Production Sharing Contract (PSC) Gross Split data
        based on available sheets.

        Returns
        -------
        PSCGrossSplitData
            An instance of PSCGrossSplitData representing the extracted data
            stored as its attributes.

        Notes
        -----
        The procedures are as follows:
        (1) Slice the data, only capture the columns that contain necessary data,
        (2) Convert the remaining NaN values to None,
        (3) Assign data to their associated attributes,
        (4) Return a new instance of PSCGrossSplitData with filled attributes.
        """
        # Step #1 - Step #2 (See 'Notes' section in the docstring)
        psc_gs_data_loaded = (
            self.data_loaded["Gross Split Config"]
            .iloc[:, 2]
            .replace(np.nan, None)
        )

        # Step #3 - Step #4 (See 'Notes' section in the docstring)
        return PSCGrossSplitData(
            field_status=psc_gs_data_loaded.iloc[0],
            field_location=psc_gs_data_loaded.iloc[1],
            reservoir_depth=psc_gs_data_loaded.iloc[2],
            infrastructure_availability=psc_gs_data_loaded.iloc[3],
            reservoir_type=psc_gs_data_loaded.iloc[4],
            co2_content=psc_gs_data_loaded.iloc[5],
            h2s_content=psc_gs_data_loaded.iloc[6],
            oil_api=psc_gs_data_loaded.iloc[7],
            domestic_content_use=psc_gs_data_loaded.iloc[8],
            production_stage=psc_gs_data_loaded.iloc[9],
            ministry_discretion_split=psc_gs_data_loaded.iloc[11],
            oil_base_split=psc_gs_data_loaded.iloc[13],
            gas_base_split=psc_gs_data_loaded.iloc[14],
            dmo_is_weighted=psc_gs_data_loaded.iloc[17],
            oil_dmo_holiday=psc_gs_data_loaded.iloc[20],
            oil_dmo_period=psc_gs_data_loaded.iloc[21],
            oil_dmo_start_production=psc_gs_data_loaded.iloc[22],
            oil_dmo_volume=psc_gs_data_loaded.iloc[23],
            oil_dmo_fee=psc_gs_data_loaded.iloc[24],
            gas_dmo_holiday=psc_gs_data_loaded.iloc[27],
            gas_dmo_period=psc_gs_data_loaded.iloc[28],
            gas_dmo_start_production=psc_gs_data_loaded.iloc[29],
            gas_dmo_volume=psc_gs_data_loaded.iloc[30],
            gas_dmo_fee=psc_gs_data_loaded.iloc[31],
        )

    def _get_psc_transition_cr_to_cr(self) -> dict:
        """
        Extract and process the associated data for PSC transition CR to CR.

        Returns
        -------
        dict
            A dictionary containing PSCCostRecoveryData objects.

        Notes
        -----
        The undertaken procedures are as follows:
        (1) Filter attribute self.sheets_loaded for sheets that contain 'Cost Recovery Config' data,
            then assigned it as local variable named 'psc_cr_to_cr_available',
        (2) Capture psc_cr_to_cr data from attribute self.data_loaded, then perform the
            necessary adjustment. Stored the result in a variable named 'psc_cr_to_cr_data_loaded',
        (3) Define a variable named 'psc_cr_to_cr_data' as a dictionary that contain two
            keys: (1) 'Cost Recovery Config', and (2) 'Cost Recovery Config (2)'. Each key is
            an instance of PSCCostRecoveryData with the associated attributes loaded from
            'psc_cr_to_cr_data_loaded',
        (4) Return the variable 'psc_cr_to_cr_data'.

        Example
        -------
        transition_cr_to_cr_data = instance._get_psc_transition_cr_to_cr()
        config_1 = transition_cr_to_cr_data['Cost Recovery Config']
        config_2 = transition_cr_to_cr_data['Cost Recovery Config (2)']
        """
        # Step #1 (See 'Notes' section in the docstring)
        psc_cr_to_cr_available = list(filter(lambda i: "Cost Recovery Config" in i, self.sheets_loaded))

        # Step #2 (See 'Notes' section in the docstring)
        psc_cr_to_cr_data_loaded = {
            key: self.data_loaded[key].replace(np.nan, None) for key in psc_cr_to_cr_available
        }

        # Step #3 - Step #4 (See 'Notes' section in the docstring)
        psc_cr_to_cr_data = {
            key: PSCCostRecoveryData(
                oil_ftp_availability=psc_cr_to_cr_data_loaded[key].iloc[0, 2],
                oil_ftp_is_shared=psc_cr_to_cr_data_loaded[key].iloc[1, 2],
                oil_ftp_portion=psc_cr_to_cr_data_loaded[key].iloc[2, 2],
                gas_ftp_availability=psc_cr_to_cr_data_loaded[key].iloc[3, 2],
                gas_ftp_is_shared=psc_cr_to_cr_data_loaded[key].iloc[4, 2],
                gas_ftp_portion=psc_cr_to_cr_data_loaded[key].iloc[5, 2],
                split_type=psc_cr_to_cr_data_loaded[key].iloc[8, 2],
                oil_ctr_pretax=psc_cr_to_cr_data_loaded[key].iloc[9, 2],
                gas_ctr_pretax=psc_cr_to_cr_data_loaded[key].iloc[10, 2],
                ic_availability=psc_cr_to_cr_data_loaded[key].iloc[13, 2],
                ic_oil=psc_cr_to_cr_data_loaded[key].iloc[14, 2],
                ic_gas=psc_cr_to_cr_data_loaded[key].iloc[15, 2],
                oil_cr_cap_rate=psc_cr_to_cr_data_loaded[key].iloc[18, 2],
                gas_cr_cap_rate=psc_cr_to_cr_data_loaded[key].iloc[19, 2],
                dmo_is_weighted=psc_cr_to_cr_data_loaded[key].iloc[22, 2],
                oil_dmo_holiday=psc_cr_to_cr_data_loaded[key].iloc[25, 2],
                oil_dmo_period=psc_cr_to_cr_data_loaded[key].iloc[26, 2],
                oil_dmo_start_production=psc_cr_to_cr_data_loaded[key].iloc[27, 2],
                oil_dmo_volume=psc_cr_to_cr_data_loaded[key].iloc[28, 2],
                oil_dmo_fee=psc_cr_to_cr_data_loaded[key].iloc[29, 2],
                gas_dmo_holiday=psc_cr_to_cr_data_loaded[key].iloc[32, 2],
                gas_dmo_period=psc_cr_to_cr_data_loaded[key].iloc[33, 2],
                gas_dmo_start_production=psc_cr_to_cr_data_loaded[key].iloc[34, 2],
                gas_dmo_volume=psc_cr_to_cr_data_loaded[key].iloc[35, 2],
                gas_dmo_fee=psc_cr_to_cr_data_loaded[key].iloc[36, 2],
                rc_split_init=psc_cr_to_cr_data_loaded[key].iloc[41:48, 1:].to_numpy(),
                icp_sliding_scale_init=psc_cr_to_cr_data_loaded[key].iloc[52:60, 1:].to_numpy(),
                indicator_rc_split_sliding_scale_init=(
                    psc_cr_to_cr_data_loaded[key].iloc[63:83, 1:3].to_numpy()
                ),
            )
            for key in psc_cr_to_cr_available
        }

        return psc_cr_to_cr_data

    def _get_psc_transition_cr_to_gs(self) -> dict:
        """
        Extract and process the associated data for PSC transition CR to GS.

        Returns
        -------
        dict
            A dictionary containing PSCCostRecoveryData and PSCGrossSplitData objects.

        Notes
        -----
        The undertaken procedures are as follows:
        (1) Load the PSC CR and PSC GS data, respectively,
        (2) Combine both data into a dictionary called 'psc_cr_to_gs_data_loaded',
        (3) Define a variable named 'psc_cr_to_gs_data' as a dictionary that contain two
            keys: (1) 'Cost Recovery Config', and (2) 'Gross Split Config'. The value of each
            key is an instance of PSCCostRecoveryData and PSCGrossSplitData, respectively, with
            the associated attributes loaded from 'psc_cr_to_gs_data_loaded',
        (4) Return the variable 'psc_cr_to_gs_data'.

        Example
        -------
        transition_cr_to_gs_data = instance._get_psc_transition_cr_to_gs()
        config_1 = transition_cr_to_gs_data['Cost Recovery Config']
        config_2 = transition_cr_to_gs_data['Gross Split Config']
        """
        # Step #1 (See 'Notes' section in the docstring)
        psc_cr_data_loaded = self.data_loaded["Cost Recovery Config"].replace(np.nan, None)
        psc_gs_data_loaded = (
            self.data_loaded["Gross Split Config"]
            .iloc[:, 2]
            .replace(np.nan, None)
        )

        # Step #2 (See 'Notes' section in the docstring)
        sheets_target = ["Cost Recovery Config", "Gross Split Config"]
        data_loaded_target = [psc_cr_data_loaded, psc_gs_data_loaded]
        psc_cr_to_gs_data_loaded = {key: data_loaded_target[i] for i, key in enumerate(sheets_target)}

        # Step #3 - Step #4 (See 'Notes' section in the docstring)
        psc_cr_to_gs_data = {
            sheets_target[0]: PSCCostRecoveryData(
                oil_ftp_availability=psc_cr_to_gs_data_loaded[sheets_target[0]].iloc[0, 2],
                oil_ftp_is_shared=psc_cr_to_gs_data_loaded[sheets_target[0]].iloc[1, 2],
                oil_ftp_portion=psc_cr_to_gs_data_loaded[sheets_target[0]].iloc[2, 2],
                gas_ftp_availability=psc_cr_to_gs_data_loaded[sheets_target[0]].iloc[3, 2],
                gas_ftp_is_shared=psc_cr_to_gs_data_loaded[sheets_target[0]].iloc[4, 2],
                gas_ftp_portion=psc_cr_to_gs_data_loaded[sheets_target[0]].iloc[5, 2],
                split_type=psc_cr_to_gs_data_loaded[sheets_target[0]].iloc[8, 2],
                oil_ctr_pretax=psc_cr_to_gs_data_loaded[sheets_target[0]].iloc[9, 2],
                gas_ctr_pretax=psc_cr_to_gs_data_loaded[sheets_target[0]].iloc[10, 2],
                ic_availability=psc_cr_to_gs_data_loaded[sheets_target[0]].iloc[13, 2],
                ic_oil=psc_cr_to_gs_data_loaded[sheets_target[0]].iloc[14, 2],
                ic_gas=psc_cr_to_gs_data_loaded[sheets_target[0]].iloc[15, 2],
                oil_cr_cap_rate=psc_cr_to_gs_data_loaded[sheets_target[0]].iloc[18, 2],
                gas_cr_cap_rate=psc_cr_to_gs_data_loaded[sheets_target[0]].iloc[19, 2],
                dmo_is_weighted=psc_cr_to_gs_data_loaded[sheets_target[0]].iloc[22, 2],
                oil_dmo_holiday=psc_cr_to_gs_data_loaded[sheets_target[0]].iloc[25, 2],
                oil_dmo_period=psc_cr_to_gs_data_loaded[sheets_target[0]].iloc[26, 2],
                oil_dmo_start_production=psc_cr_to_gs_data_loaded[sheets_target[0]].iloc[27, 2],
                oil_dmo_volume=psc_cr_to_gs_data_loaded[sheets_target[0]].iloc[28, 2],
                oil_dmo_fee=psc_cr_to_gs_data_loaded[sheets_target[0]].iloc[29, 2],
                gas_dmo_holiday=psc_cr_to_gs_data_loaded[sheets_target[0]].iloc[32, 2],
                gas_dmo_period=psc_cr_to_gs_data_loaded[sheets_target[0]].iloc[33, 2],
                gas_dmo_start_production=psc_cr_to_gs_data_loaded[sheets_target[0]].iloc[34, 2],
                gas_dmo_volume=psc_cr_to_gs_data_loaded[sheets_target[0]].iloc[35, 2],
                gas_dmo_fee=psc_cr_to_gs_data_loaded[sheets_target[0]].iloc[36, 2],
                rc_split_init=psc_cr_to_gs_data_loaded[sheets_target[0]].iloc[41:48, 1:].to_numpy(),
                icp_sliding_scale_init=(
                    psc_cr_to_gs_data_loaded[sheets_target[0]].iloc[52:60, 1:].to_numpy()
                ),
                indicator_rc_split_sliding_scale_init=(
                    psc_cr_to_gs_data_loaded[sheets_target[0]].iloc[63:83, 1:3].to_numpy()
                ),
            ),
            sheets_target[1]: PSCGrossSplitData(
                field_status=psc_cr_to_gs_data_loaded[sheets_target[1]].iloc[0],
                field_location=psc_cr_to_gs_data_loaded[sheets_target[1]].iloc[1],
                reservoir_depth=psc_cr_to_gs_data_loaded[sheets_target[1]].iloc[2],
                infrastructure_availability=psc_cr_to_gs_data_loaded[sheets_target[1]].iloc[3],
                reservoir_type=psc_cr_to_gs_data_loaded[sheets_target[1]].iloc[4],
                co2_content=psc_cr_to_gs_data_loaded[sheets_target[1]].iloc[5],
                h2s_content=psc_cr_to_gs_data_loaded[sheets_target[1]].iloc[6],
                oil_api=psc_cr_to_gs_data_loaded[sheets_target[1]].iloc[7],
                domestic_content_use=psc_cr_to_gs_data_loaded[sheets_target[1]].iloc[8],
                production_stage=psc_cr_to_gs_data_loaded[sheets_target[1]].iloc[9],
                ministry_discretion_split=psc_cr_to_gs_data_loaded[sheets_target[1]].iloc[11],
                oil_base_split=psc_cr_to_gs_data_loaded[sheets_target[1]].iloc[13],
                gas_base_split=psc_cr_to_gs_data_loaded[sheets_target[1]].iloc[14],
                dmo_is_weighted=psc_cr_to_gs_data_loaded[sheets_target[1]].iloc[17],
                oil_dmo_holiday=psc_cr_to_gs_data_loaded[sheets_target[1]].iloc[20],
                oil_dmo_period=psc_cr_to_gs_data_loaded[sheets_target[1]].iloc[21],
                oil_dmo_start_production=psc_cr_to_gs_data_loaded[sheets_target[1]].iloc[22],
                oil_dmo_volume=psc_cr_to_gs_data_loaded[sheets_target[1]].iloc[23],
                oil_dmo_fee=psc_cr_to_gs_data_loaded[sheets_target[1]].iloc[24],
                gas_dmo_holiday=psc_cr_to_gs_data_loaded[sheets_target[1]].iloc[27],
                gas_dmo_period=psc_cr_to_gs_data_loaded[sheets_target[1]].iloc[28],
                gas_dmo_start_production=psc_cr_to_gs_data_loaded[sheets_target[1]].iloc[29],
                gas_dmo_volume=psc_cr_to_gs_data_loaded[sheets_target[1]].iloc[30],
                gas_dmo_fee=psc_cr_to_gs_data_loaded[sheets_target[1]].iloc[31],
            )
        }

        return psc_cr_to_gs_data

    def _get_psc_transition_gs_to_gs(self) -> dict:
        """
        Extract and process the associated data for PSC transition GS to GS.

        Returns
        -------
        dict
            A dictionary containing PSCGrossSplitData objects.

        Notes
        -----
        The undertaken procedures are as follows:
        (1) Filter attribute self.sheets_loaded for sheets that contain 'Gross Split Config' data,
            then assigned it as local variable named 'psc_gs_to_gs_available',
        (2) Capture psc_gs_to_gs data from attribute self.data_loaded, then perform the
            necessary adjustment. Stored the result in a variable named 'psc_gs_to_gs_data_loaded',
        (3) Define a variable named 'psc_gs_to_gs_data' as a dictionary that contain two
            keys: (1) 'Gross Split Config', and (2) 'Gross Split Config (2)'. Each keys is
            an instance of PSCGrossSplitData with the associated attributes loaded from
            'psc_gs_to_gs_data_loaded',
        (4) Return the variable 'psc_gs_to_gs_data'.

        Example
        -------
        transition_gs_to_gs_data = instance._get_psc_transition_gs_to_gs()
        config_1 = transition_gs_to_gs_data['Gross Split Config']
        config_2 = transition_gs_to_gs_data['Gross Split Config (2)']
        """
        # Step #1 (See 'Notes' section in the docstring)
        psc_gs_to_gs_available = list(filter(lambda i: "Gross Split Config" in i, self.sheets_loaded))

        # Step #2 (See 'Notes' section in the docstring)
        psc_gs_to_gs_data_loaded = {
            key: self.data_loaded[key].iloc[:, 2].replace(np.nan, None)
            for key in psc_gs_to_gs_available
        }

        # Step #3 - Step #4 (See 'Notes' section in the docstring)
        psc_gs_to_gs_data = {
            key: PSCGrossSplitData(
                field_status=psc_gs_to_gs_data_loaded[key].iloc[0],
                field_location=psc_gs_to_gs_data_loaded[key].iloc[1],
                reservoir_depth=psc_gs_to_gs_data_loaded[key].iloc[2],
                infrastructure_availability=psc_gs_to_gs_data_loaded[key].iloc[3],
                reservoir_type=psc_gs_to_gs_data_loaded[key].iloc[4],
                co2_content=psc_gs_to_gs_data_loaded[key].iloc[5],
                h2s_content=psc_gs_to_gs_data_loaded[key].iloc[6],
                oil_api=psc_gs_to_gs_data_loaded[key].iloc[7],
                domestic_content_use=psc_gs_to_gs_data_loaded[key].iloc[8],
                production_stage=psc_gs_to_gs_data_loaded[key].iloc[9],
                ministry_discretion_split=psc_gs_to_gs_data_loaded[key].iloc[11],
                oil_base_split=psc_gs_to_gs_data_loaded[key].iloc[13],
                gas_base_split=psc_gs_to_gs_data_loaded[key].iloc[14],
                dmo_is_weighted=psc_gs_to_gs_data_loaded[key].iloc[17],
                oil_dmo_holiday=psc_gs_to_gs_data_loaded[key].iloc[20],
                oil_dmo_period=psc_gs_to_gs_data_loaded[key].iloc[21],
                oil_dmo_start_production=psc_gs_to_gs_data_loaded[key].iloc[22],
                oil_dmo_volume=psc_gs_to_gs_data_loaded[key].iloc[23],
                oil_dmo_fee=psc_gs_to_gs_data_loaded[key].iloc[24],
                gas_dmo_holiday=psc_gs_to_gs_data_loaded[key].iloc[27],
                gas_dmo_period=psc_gs_to_gs_data_loaded[key].iloc[28],
                gas_dmo_start_production=psc_gs_to_gs_data_loaded[key].iloc[29],
                gas_dmo_volume=psc_gs_to_gs_data_loaded[key].iloc[30],
                gas_dmo_fee=psc_gs_to_gs_data_loaded[key].iloc[31],
            )
            for key in psc_gs_to_gs_available
        }

        return psc_gs_to_gs_data

    def _get_psc_transition_gs_to_cr(self) -> dict:
        """
        Extract and process the associated data for PSC transition GS to CR.

        Returns
        -------
        dict
            A dictionary containing PSCGrossSplitData and PSCCostRecoveryData objects.

        Notes
        -----
        The undertaken procedures are as follows:
        (1) Load the PSC GS and PSC CR data, respectively,
        (2) Combine both data into a dictionary called 'psc_gs_to_cr_data_loaded',
        (3) Define a variable named 'psc_gs_to_cr_data' as a dictionary that contain two
            keys: (1) 'Gross Split Config', and (2) 'Cost Recovery Config'. The value of each
            key is an instance of PSCGrossSplitData and PSCCostRecoveryData, respectively, with
            the associated attributes loaded from 'psc_gs_to_cr_data_loaded',
        (4) Return the variable 'psc_gs_to_cr_data'.

        Example
        -------
        transition_gs_to_cr_data = instance._get_psc_transition_gs_to_cr()
        config_1 = transition_gs_to_cr_data['Gross Split Config']
        config_2 = transition_gs_to_cr_data['Cost Recovery Config']
        """
        # Step #1 (See 'Notes' section in the docstring)
        psc_gs_data_loaded = (
            self.data_loaded["Gross Split Config"]
            .iloc[:, 2]
            .replace(np.nan, None)
        )
        psc_cr_data_loaded = self.data_loaded["Cost Recovery Config"].replace(np.nan, None)

        # Step #2 (See 'Notes' section in the docstring)
        sheets_target = ["Gross Split Config", "Cost Recovery Config"]
        data_loaded_target = [psc_gs_data_loaded, psc_cr_data_loaded]
        psc_gs_to_cr_data_loaded = {key: data_loaded_target[i] for i, key in enumerate(sheets_target)}

        # Step #3 - Step #4 (See 'Notes' section in the docstring)
        psc_gs_to_cr_data = {
            sheets_target[0]: PSCGrossSplitData(
                field_status=psc_gs_to_cr_data_loaded[sheets_target[0]].iloc[0],
                field_location=psc_gs_to_cr_data_loaded[sheets_target[0]].iloc[1],
                reservoir_depth=psc_gs_to_cr_data_loaded[sheets_target[0]].iloc[2],
                infrastructure_availability=psc_gs_to_cr_data_loaded[sheets_target[0]].iloc[3],
                reservoir_type=psc_gs_to_cr_data_loaded[sheets_target[0]].iloc[4],
                co2_content=psc_gs_to_cr_data_loaded[sheets_target[0]].iloc[5],
                h2s_content=psc_gs_to_cr_data_loaded[sheets_target[0]].iloc[6],
                oil_api=psc_gs_to_cr_data_loaded[sheets_target[0]].iloc[7],
                domestic_content_use=psc_gs_to_cr_data_loaded[sheets_target[0]].iloc[8],
                production_stage=psc_gs_to_cr_data_loaded[sheets_target[0]].iloc[9],
                ministry_discretion_split=psc_gs_to_cr_data_loaded[sheets_target[0]].iloc[11],
                oil_base_split=psc_gs_to_cr_data_loaded[sheets_target[0]].iloc[13],
                gas_base_split=psc_gs_to_cr_data_loaded[sheets_target[0]].iloc[14],
                dmo_is_weighted=psc_gs_to_cr_data_loaded[sheets_target[0]].iloc[17],
                oil_dmo_holiday=psc_gs_to_cr_data_loaded[sheets_target[0]].iloc[20],
                oil_dmo_period=psc_gs_to_cr_data_loaded[sheets_target[0]].iloc[21],
                oil_dmo_start_production=psc_gs_to_cr_data_loaded[sheets_target[0]].iloc[22],
                oil_dmo_volume=psc_gs_to_cr_data_loaded[sheets_target[0]].iloc[23],
                oil_dmo_fee=psc_gs_to_cr_data_loaded[sheets_target[0]].iloc[24],
                gas_dmo_holiday=psc_gs_to_cr_data_loaded[sheets_target[0]].iloc[27],
                gas_dmo_period=psc_gs_to_cr_data_loaded[sheets_target[0]].iloc[28],
                gas_dmo_start_production=psc_gs_to_cr_data_loaded[sheets_target[0]].iloc[29],
                gas_dmo_volume=psc_gs_to_cr_data_loaded[sheets_target[0]].iloc[30],
                gas_dmo_fee=psc_gs_to_cr_data_loaded[sheets_target[0]].iloc[31],
            ),
            sheets_target[1]: PSCCostRecoveryData(
                oil_ftp_availability=psc_gs_to_cr_data_loaded[sheets_target[1]].iloc[0, 2],
                oil_ftp_is_shared=psc_gs_to_cr_data_loaded[sheets_target[1]].iloc[1, 2],
                oil_ftp_portion=psc_gs_to_cr_data_loaded[sheets_target[1]].iloc[2, 2],
                gas_ftp_availability=psc_gs_to_cr_data_loaded[sheets_target[1]].iloc[3, 2],
                gas_ftp_is_shared=psc_gs_to_cr_data_loaded[sheets_target[1]].iloc[4, 2],
                gas_ftp_portion=psc_gs_to_cr_data_loaded[sheets_target[1]].iloc[5, 2],
                split_type=psc_gs_to_cr_data_loaded[sheets_target[1]].iloc[8, 2],
                oil_ctr_pretax=psc_gs_to_cr_data_loaded[sheets_target[1]].iloc[9, 2],
                gas_ctr_pretax=psc_gs_to_cr_data_loaded[sheets_target[1]].iloc[10, 2],
                ic_availability=psc_gs_to_cr_data_loaded[sheets_target[1]].iloc[13, 2],
                ic_oil=psc_gs_to_cr_data_loaded[sheets_target[1]].iloc[14, 2],
                ic_gas=psc_gs_to_cr_data_loaded[sheets_target[1]].iloc[15, 2],
                oil_cr_cap_rate=psc_gs_to_cr_data_loaded[sheets_target[1]].iloc[18, 2],
                gas_cr_cap_rate=psc_gs_to_cr_data_loaded[sheets_target[1]].iloc[19, 2],
                dmo_is_weighted=psc_gs_to_cr_data_loaded[sheets_target[1]].iloc[22, 2],
                oil_dmo_holiday=psc_gs_to_cr_data_loaded[sheets_target[1]].iloc[25, 2],
                oil_dmo_period=psc_gs_to_cr_data_loaded[sheets_target[1]].iloc[26, 2],
                oil_dmo_start_production=psc_gs_to_cr_data_loaded[sheets_target[1]].iloc[27, 2],
                oil_dmo_volume=psc_gs_to_cr_data_loaded[sheets_target[1]].iloc[28, 2],
                oil_dmo_fee=psc_gs_to_cr_data_loaded[sheets_target[1]].iloc[29, 2],
                gas_dmo_holiday=psc_gs_to_cr_data_loaded[sheets_target[1]].iloc[32, 2],
                gas_dmo_period=psc_gs_to_cr_data_loaded[sheets_target[1]].iloc[33, 2],
                gas_dmo_start_production=psc_gs_to_cr_data_loaded[sheets_target[1]].iloc[34, 2],
                gas_dmo_volume=psc_gs_to_cr_data_loaded[sheets_target[1]].iloc[35, 2],
                gas_dmo_fee=psc_gs_to_cr_data_loaded[sheets_target[1]].iloc[36, 2],
                rc_split_init=psc_gs_to_cr_data_loaded[sheets_target[1]].iloc[41:48, 1:].to_numpy(),
                icp_sliding_scale_init=(
                    psc_gs_to_cr_data_loaded[sheets_target[1]].iloc[52:60, 1:].to_numpy()
                ),
                indicator_rc_split_sliding_scale_init=(
                    psc_gs_to_cr_data_loaded[sheets_target[1]].iloc[63:83, 1:3].to_numpy()
                ),
            )
        }

        return psc_gs_to_cr_data

    def _get_sensitivity_data(self) -> SensitivityData:
        """
        Extract and process the associated data for sensitivity analysis.

        Returns
        -------
        SensitivityData
            An instance of SensitivityData.

        Notes
        -----
        The undertaken procedures are as follows:
        (1) Extract sensitivity data from 'self.data_loaded',
        (2) Create a new instance of SensitivityData with the necessary information
            stored as the corresponding attributes.
        """
        sensitivity_data_loaded = self.data_loaded["Sensitivity"].iloc[0:11, 1:3]

        return SensitivityData(
            parameter=sensitivity_data_loaded.iloc[6:, 0].to_numpy().tolist(),
            percentage_min=sensitivity_data_loaded.iloc[0, 1],
            percentage_max=sensitivity_data_loaded.iloc[1, 1],
            step=sensitivity_data_loaded.iloc[2, 1],
        )

    def _get_montecarlo_data(self) -> MonteCarloData:
        """
        Extract and process the associated data for montecarlo analysis.

        Returns
        -------
        MonteCarloData
            An instance of MonteCarloData.

        Notes
        -----
        The undertaken procedures are as follows:
        (1) Extract montecarlo data from 'self.data_loaded',
        (2) Create a new instance of MonteCarloData with the necessary information
            stored as the corresponding attributes.
        """
        mc_data_loaded = self.data_loaded["Uncertainty"].iloc[1:10, np.r_[1:3, 4:8]]

        return MonteCarloData(
            run_number=mc_data_loaded.iloc[0, 1],
            parameter=mc_data_loaded.iloc[3:8, 0].to_numpy().tolist(),
            distribution=mc_data_loaded.iloc[3:8, 1].to_numpy().tolist(),
            min_values=mc_data_loaded.iloc[3:8, 2].to_numpy(),
            mean_values=mc_data_loaded.iloc[3:8, 3].to_numpy(),
            max_values=mc_data_loaded.iloc[3:8, 4].to_numpy(),
            std_dev=mc_data_loaded.iloc[3:8, 5].to_numpy(),
        )

    def _get_optimization_data(self) -> OptimizationData:
        """
        Extract and process the associated data for optimization study.

        Returns
        -------
        OptimizationData
            An instance of OptimizationData.

        Notes
        -----
        The undertaken operations are as follows:
        (1) Extract optimization data from 'self.data_loaded' and perform the necessary adjustments,
        (2) Configure the data associated with PSC Cost Recovery optimization,
        (3) Configure the data associated with PSC Gross Split optimization,
        (4) Configure the data associated with objective function,
        (5) Create an instance of OptimizationData and store the above data as its attributes.
        """
        # Step #1 (See 'Notes' section in the docstring)
        optimization_data_loaded = (
            self.data_loaded["Optimization"]
            .iloc[:, np.r_[1, 4, 6:8, 3, 5]]
            # .replace(np.nan, None)
        )

        # Step #2 (See 'Notes' section in the docstring)
        data_cr_init = {
            key: optimization_data_loaded.iloc[3:13, i].to_numpy()
            for i, key in enumerate(["parameter", "priority", "min", "max"])
        }

        # Step #3 (See 'Notes' section in the docstring)
        data_gs_init = {
            key: optimization_data_loaded.iloc[15:, i].to_numpy()
            for i, key in enumerate(["parameter", "priority", "min", "max"])
        }

        # Step #4 (See 'Notes' section in the docstring)
        target = {
            key: optimization_data_loaded.iloc[0, i + 4]
            for i, key in enumerate(["parameter", "value"])
        }

        # Step #5 (See 'Notes' section in the docstring)
        return OptimizationData(
            target_init=target,
            data_cr_init=data_cr_init,
            data_gs_init=data_gs_init,
            type_of_contract=self.general_config_data.type_of_contract,
        )

    def prepare_data(self) -> None:
        """
        Method to prepare data of Spreadsheet class.

        Reads and prepare data from a target Excel file and fills in attributes
        associated with various categories, including configuration, lifting, cost,
        contract, and additional functionality.

        Returns
        -------
        None
            This method modifies the instance attributes of Spreadsheet class.

        Notes
        -----
        This method serves as a comprehensive data preparation step, calling specific
        private methods to obtain and assign data based on their category.

        The attributes include general configuration data, fiscal configuration data,
        lifting data for oil, gas, LPG propane, LPG butane, sulfur, electricity, and CO2, cost data
        including tangible and intangible costs, opex, and ASR costs, contract data including
        PSC Cost Recovery (CR), Gross Split (GS), and transition cases, as well as additional
        functionality data such as sensitivity, Monte Carlo, and optimization data.
        """
        # Read data from a target Excel file
        self.read_from_excel()

        # Fill in the attributes associated with config data
        self.general_config_data = self._get_general_config_data()
        self.fiscal_config_data = self._get_fiscal_config_data()

        # Fill in the attributes associated with lifting data
        self.oil_lifting_data = self._get_oil_lifting_data()
        self.gas_lifting_data = self._get_gas_lifting_data()
        self.lpg_propane_lifting_data = self._get_lpg_propane_lifting_data()
        self.lpg_butane_lifting_data = self._get_lpg_butane_lifting_data()
        self.sulfur_lifting_data = self._get_sulfur_lifting_data()
        self.electricity_lifting_data = self._get_electricity_lifting_data()
        self.co2_lifting_data = self._get_co2_lifting_data()

        # Fill in the attributes associated with cost data
        self.capital_cost_data = self._get_capital_cost_data()
        self.intangible_cost_data = self._get_intangible_cost_data()
        self.opex_data = self._get_opex_data()
        self.asr_cost_data = self._get_asr_cost_data()

        # Fill in the attributes associated with contract data
        self.psc_cr_data = self._get_psc_cr_data()
        self.psc_gs_data = self._get_psc_gs_data()
        self.psc_transition_cr_to_cr = self._get_psc_transition_cr_to_cr()
        self.psc_transition_cr_to_gs = self._get_psc_transition_cr_to_gs()
        self.psc_transition_gs_to_gs = self._get_psc_transition_gs_to_gs()
        self.psc_transition_gs_to_cr = self._get_psc_transition_gs_to_cr()

        # Fill in the attributes associated with sensitivity and optimization
        self.optimization_data = self._get_optimization_data()
        self.sensitivity_data = self._get_sensitivity_data()
        self.montecarlo_data = self._get_montecarlo_data()



"""
A unit testing for depreciation module.
"""

import numpy as np
from pyscnomics.econ.depreciation import (
    straight_line_depreciation_rate,
    declining_balance_depreciation_rate,
    psc_declining_balance_depreciation_rate,
    straight_line_book_value,
    declining_balance_book_value,
    psc_declining_balance_book_value,
)


def test_straight_line_depreciation():
    """A unit testing for straight line depreciation"""

    # Expected results
    depreciationCharge1 = [35.0, 35.0, 35.0, 35.0, 35.0]
    depreciationCharge2 = [35.0, 35.0, 35.0, 35.0, 35.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    depreciationCharge3 = [33.33333333, 33.33333333, 33.33333333]

    bookValue1 = [165.0, 130.0, 95.0, 60.0, 25.0]
    bookValue2 = [165.0, 130.0, 95.0, 60.0, 25.0, 25.0, 25.0, 25.0, 25.0, 25.0]
    bookValue3 = [66.66666666, 33.33333333, 0.0]

    # Calculated results
    calcDepreciationCharge1 = straight_line_depreciation_rate(
        cost=200, salvage_value=25, useful_life=5
    )

    calcDepreciationCharge2 = straight_line_depreciation_rate(
        cost=200, salvage_value=25, useful_life=5, depreciation_len=10
    )

    calcDepreciationCharge3 = straight_line_depreciation_rate(
        cost=100, salvage_value=0, useful_life=3
    )

    calcBookValue1 = straight_line_book_value(cost=200, salvage_value=25, useful_life=5)

    calcBookValue2 = straight_line_book_value(
        cost=200, salvage_value=25, useful_life=5, depreciation_len=10
    )

    calcBookValue3 = straight_line_book_value(cost=100, salvage_value=0, useful_life=3)

    # Execute testing (expected = calculated)
    np.testing.assert_allclose(depreciationCharge1, calcDepreciationCharge1)
    np.testing.assert_allclose(depreciationCharge2, calcDepreciationCharge2)
    np.testing.assert_allclose(depreciationCharge3, calcDepreciationCharge3)
    np.testing.assert_allclose(bookValue1, calcBookValue1)
    np.testing.assert_allclose(bookValue2, calcBookValue2)
    np.testing.assert_allclose(bookValue3, calcBookValue3)


def test_decline_balance_depreciation():
    """
    A unit testing for declining balance depreciation.

    Notes
    -----
    Case 1: salvage value is 0; useful life is 5; project duration is 10 years.
    Case 2: salvage value is 50; useful life is 5;; project duration is 10 years.

    In both cases, the value of depreciation charge at the t = useful_life has to
    account for the remaining cost, namely depreciation charge at t = useful life
    has to be equal to (cost - salvage_value - sum(depreciation_charge)).

    As a result, in the final array of depreciation charge, the following condition has to apply:
    (cost - salvage_value - sum(depreciation_charge)) = 0.
    """

    cost = 100.0
    salvage1 = 0.0
    salvage2 = 50.0
    tolerance = 1e-12

    # Expected results
    depre1 = [20, 16, 12.8, 10.24, 40.96, 0, 0, 0, 0, 0]
    book1 = [80, 64, 51.2, 40.96, 0, 0, 0, 0, 0, 0]

    depre2 = [20.0, 16.0, 12.8, 1.2, 0, 0, 0, 0, 0, 0]
    book2 = [80.0, 64.0, 51.2, 50.0, 50.0, 50.0, 50.0, 50.0, 50.0, 50.0]

    # Calculated results (expected == calculated)
    calc_depre1 = declining_balance_depreciation_rate(
        cost=cost,
        salvage_value=salvage1,
        useful_life=5,
        decline_factor=1,
        depreciation_len=10,
    )

    calc_book1 = declining_balance_book_value(
        cost=cost,
        salvage_value=salvage1,
        useful_life=5,
        decline_factor=1,
        depreciation_len=10,
    )

    calc_depre2 = declining_balance_depreciation_rate(
        cost=cost,
        salvage_value=salvage2,
        useful_life=5,
        decline_factor=1,
        depreciation_len=10,
    )

    calc_book2 = declining_balance_book_value(
        cost=cost,
        salvage_value=salvage2,
        useful_life=5,
        decline_factor=1,
        depreciation_len=10,
    )

    # Execute testing
    np.testing.assert_allclose(depre1, calc_depre1)
    np.testing.assert_allclose(
        cost - salvage1 - np.sum(calc_depre1, keepdims=True),
        np.array([0.0]),
        atol=tolerance,
    )
    np.testing.assert_allclose(book1, calc_book1)

    np.testing.assert_allclose(depre2, calc_depre2)
    np.testing.assert_allclose(
        cost - salvage2 - np.sum(calc_depre2, keepdims=True),
        np.array([0.0]),
        atol=tolerance,
    )
    np.testing.assert_allclose(book2, calc_book2)


def test_double_decline_balance_depreciation():
    """
    A unit testing for double declining balance depreciation.

    Notes
    -----
    Case 1: salvage value is 0; useful life is 5; project duration is 10 years.
    Case 2: salvage value is 50; useful life is 5;; project duration is 10 years.

    In both cases, the value of depreciation charge at the t = useful_life has to
    account for the remaining cost, namely depreciation charge at t = useful life
    has to be equal to (cost - salvage_value - sum(depreciation_charge)).

    As a result, in the final array of depreciation charge, the following condition has to apply:
    (cost - salvage_value - sum(depreciation_charge)) = 0.
    """

    cost = 100.0
    salvage1 = 0.0
    salvage2 = 50.0
    tolerance = 1e-12

    # Expected results
    depre1 = [40.0, 24.0, 14.4, 8.64, 12.96, 0.0, 0.0, 0.0, 0.0, 0.0]
    book1 = [60.0, 36.0, 21.6, 12.96, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]

    depre2 = [40.0, 10.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    book2 = [60.0, 50.0, 50.0, 50.0, 50.0, 50.0, 50.0, 50.0, 50.0, 50.0]

    # Calculated results
    calc_depre1 = declining_balance_depreciation_rate(
        cost=cost,
        salvage_value=salvage1,
        useful_life=5,
        decline_factor=2,
        depreciation_len=10,
    )

    calc_book1 = declining_balance_book_value(
        cost=cost,
        salvage_value=salvage1,
        useful_life=5,
        decline_factor=2,
        depreciation_len=10,
    )

    calc_depre2 = declining_balance_depreciation_rate(
        cost=cost,
        salvage_value=salvage2,
        useful_life=5,
        decline_factor=2,
        depreciation_len=10,
    )

    calc_book2 = declining_balance_book_value(
        cost=cost,
        salvage_value=salvage2,
        useful_life=5,
        decline_factor=2,
        depreciation_len=10,
    )

    # Execute testing (expected == calculated)
    np.testing.assert_allclose(depre1, calc_depre1)
    np.testing.assert_allclose(
        cost - salvage1 - np.sum(calc_depre1, keepdims=True),
        np.array([0.0]),
        atol=tolerance,
    )
    np.testing.assert_allclose(book1, calc_book1)

    np.testing.assert_allclose(depre2, calc_depre2)
    np.testing.assert_allclose(
        cost - salvage2 - np.sum(calc_depre2, keepdims=True),
        np.array([0.0]),
        atol=tolerance,
    )
    np.testing.assert_allclose(book2, calc_book2)


def test_psc_decline_balance_depreciation():
    """
    A unit testing for psc declining balance depreciation.

    Notes
    -----
    Case 1: useful_life = 5; project_duration = 10
    Case 2: useful_life = 3; project_duration = 10
    """

    cost = 100
    useful_life1 = 5
    useful_life2 = 3
    tolerance = 1e-12

    # Expected results
    depre1 = [50, 25, 12.5, 6.25, 6.25, 0, 0, 0, 0, 0]
    book1 = [50, 25, 12.5, 6.25, 0, 0, 0, 0, 0, 0]

    depre2 = [50, 25, 25, 0, 0, 0, 0, 0, 0, 0]
    book2 = [50, 25, 0, 0, 0, 0, 0, 0, 0, 0]

    # Calculated results
    calc_depre1 = psc_declining_balance_depreciation_rate(
        cost=cost,
        useful_life=useful_life1,
        depreciation_factor=0.5,
        depreciation_len=10,
    )

    calc_book1 = psc_declining_balance_book_value(
        cost=cost,
        useful_life=useful_life1,
        depreciation_factor=0.5,
        depreciation_len=10,
    )

    calc_depre2 = psc_declining_balance_depreciation_rate(
        cost=cost,
        useful_life=useful_life2,
        depreciation_factor=0.5,
        depreciation_len=10,
    )

    calc_book2 = psc_declining_balance_book_value(
        cost=cost,
        useful_life=useful_life2,
        depreciation_factor=0.5,
        depreciation_len=10,
    )

    # Execute testing (expected == calculated)
    np.testing.assert_allclose(depre1, calc_depre1)
    np.testing.assert_allclose(
        cost - np.sum(calc_depre1), np.array([0]), atol=tolerance
    )
    np.testing.assert_allclose(book1, calc_book1)

    np.testing.assert_allclose(depre2, calc_depre2)
    np.testing.assert_allclose(
        cost - np.sum(calc_depre2), np.array([0]), atol=tolerance
    )
    np.testing.assert_allclose(book2, calc_book2)



"""
This file is utilized to be the adapter of the router into the core codes.
"""
import numpy as np
import pandas as pd
from datetime import datetime

from pyscnomics.contracts.project import BaseProject
from pyscnomics.contracts.costrecovery import CostRecovery
from pyscnomics.contracts.grossplit import GrossSplit
from pyscnomics.contracts.transition import Transition
from pyscnomics.tools.summary import get_summary
from pyscnomics.tools.table import get_table
from pyscnomics.optimize.optimization import optimize_psc
from pyscnomics.optimize.optimization_transition import optimize_psc_core as optimize_psc_trans
from pyscnomics.econ.selection import OptimizationParameter, FluidType
from pyscnomics.tools.ltp import oil_ltp_predict, gas_ltp_predict
from pyscnomics.tools.rpd import RPDModel
from pyscnomics.api.converter import (convert_str_to_date,
                                      convert_list_to_array_float_or_array,
                                      convert_dict_to_lifting,
                                      convert_dict_to_capital,
                                      convert_dict_to_intangible,
                                      convert_dict_to_opex,
                                      convert_dict_to_asr,
                                      convert_dict_to_lbt,
                                      convert_dict_to_cost_of_sales,
                                      convert_list_to_array_float,
                                      convert_list_to_array_float_or_array_or_none,
                                      convert_str_to_taxsplit,
                                      convert_str_to_npvmode,
                                      convert_str_to_discountingmode,
                                      convert_str_to_otherrevenue,
                                      convert_str_to_taxregime,
                                      convert_str_to_ftptaxregime,
                                      convert_str_to_depremethod,
                                      convert_str_to_inflationappliedto,
                                      convert_summary_to_dict,
                                      convert_str_to_optimization_parameters,
                                      convert_str_to_optimization_targetparameter,
                                      convert_grosssplitregime_to_enum,
                                      convert_to_float,
                                      read_fluid_type,
                                      convert_to_method_limit)
from pyscnomics.econ.limit import econ_limit


class ContractException(Exception):
    """Exception to be raised if contract type is misused"""

    pass


class LTPModelException(Exception):
    """ Exception to be raised for an incorrect LTP configurations """

    pass


class RDPModelException(Exception):
    """ Exception to be raised for an incorrect RDP configurations """

    pass


def get_setup_dict(data: dict) -> tuple:
    """
    Function to get conversion of the setup input from dictionary into acceptable core engine data format.

    Parameters
    ----------
    data: dict
        The dictionary of the data input

    Returns
    -------
    start_date: date
        The start date of the project.
    end_date: date
        The end date of the project.
    oil_onstream_date: date
        The oil onstream date.
    gas_onstream_date: date
        The gas onstream date.
    lifting: Lifting
        The lifting of the project, in Lifting Dataclass format.
    capital: Tangible
        The capital cost of the project, in Tangible Dataclass format.
    intangible: Intangible
        The intangible cost of the project, in Intangible Dataclass format.
    opex: OPEX
        The opex cost of the project, in OPEX Dataclass format.
    lbt: LBT
        The land and building tax of the project, in LBT Dataclass format.
    cost_of_sales: CostOfSales
        The opex cost of the project, in CostOfSales Dataclass format.
    asr: ASR
        The asr cost of the project, in ASR Dataclass format.

    """
    # Parsing the contract setup into each corresponding variables
    start_date = convert_str_to_date(str_object=data['setup']['start_date'])
    end_date = convert_str_to_date(str_object=data['setup']['end_date'])
    oil_onstream_date = convert_str_to_date(str_object=data['setup']['oil_onstream_date'])
    gas_onstream_date = convert_str_to_date(str_object=data['setup']['gas_onstream_date'])
    lifting = convert_dict_to_lifting(data_raw=data)
    capital = convert_dict_to_capital(data_raw=data['capital'])
    intangible = convert_dict_to_intangible(data_raw=data['intangible'])
    opex = convert_dict_to_opex(data_raw=data['opex'])
    asr = convert_dict_to_asr(data_raw=data['asr'])
    lbt = convert_dict_to_lbt(data_raw=data['lbt'])
    cost_of_sales = convert_dict_to_cost_of_sales(data_raw=data['cost_of_sales'])
    return start_date, end_date, oil_onstream_date, gas_onstream_date, lifting, capital, intangible, opex, asr, lbt, cost_of_sales


def get_summary_dict(data: dict) -> dict:
    """
    Function to get the summary arguments from the dictionary data input.
    Parameters
    ----------
    data: dict
        The dictionary of the data input

    Returns
    -------
    summary_arguments_dict: dict
        The summary argument in the core engine acceptable format.



    """
    # Filling the argument with the input data
    reference_year = data['summary_arguments']['reference_year']
    inflation_rate = data['summary_arguments']['inflation_rate']
    discount_rate = data['summary_arguments']['discount_rate']
    npv_mode = convert_str_to_npvmode(str_object=data['summary_arguments']['npv_mode'])
    discounting_mode = convert_str_to_discountingmode(str_object=data['summary_arguments']['discounting_mode'])
    profitability_discounted = data['summary_arguments']['profitability_discounted']

    summary_arguments_dict = {
        'reference_year': reference_year,
        'inflation_rate': inflation_rate,
        'discount_rate': discount_rate,
        'npv_mode': npv_mode,
        'discounting_mode': discounting_mode,
        'profitability_discounted': profitability_discounted
    }

    return summary_arguments_dict


def get_summary_object(data: dict, contract: CostRecovery | GrossSplit | Transition) -> (dict, dict):
    """
    The function to get the summary dictionary object from the data and contract input.

    Parameters
    ----------
    data: dict
        The dictionary of the data input
    contract: CostRecovery | GrossSplit | Transition
        The contract object.

    Returns
    -------
    summary: dict
        The summary of the contract ini dictionary format.

    summary_arguments_dict: dict
        The summary arguments used in retrieving the summary of the contract.

    """
    if contract is Transition:
        summary_arguments_dict = get_summary_dict(data=data)
        summary_arguments_dict['contract'] = contract
        summary = get_summary(**summary_arguments_dict)

    else:
        summary_arguments_dict = get_summary_dict(data=data)
        summary_arguments_dict['contract'] = contract
        summary = get_summary(**summary_arguments_dict)

    return summary, summary_arguments_dict


def get_costrecovery(data: dict, summary_result: bool = True):
    """
    The function to get the Summary, Cost Recovery object, contract arguments, and summary arguments used.

    Parameters
    ----------
    data: dict
        The dictionary of the data input.
    summary_result: bool
        The condition if the summary result will be generated or not.

    Returns
    -------
    summary_skk: dict
        The executive summary of the contract.
    contract:
        The Cost Recovery contract object.
    contract_arguments_dict: dict
        The contract arguments used in running the contract calculation.
    summary_arguments_dic: dict
        The summary arguments used in retrieving the executive summary of the contract.

    """
    start_date, end_date, oil_onstream_date, gas_onstream_date, lifting, tangible, intangible, opex, asr, lbt, cost_of_sales = get_setup_dict(data=data)

    contract = CostRecovery(
        start_date=start_date,
        end_date=end_date,
        oil_onstream_date=oil_onstream_date,
        gas_onstream_date=gas_onstream_date,
        lifting=lifting,
        capital_cost=tangible,
        intangible_cost=intangible,
        opex=opex,
        asr_cost=asr,
        lbt_cost=lbt,
        cost_of_sales=cost_of_sales,
        oil_ftp_is_available=data['costrecovery']['oil_ftp_is_available'],
        oil_ftp_is_shared=data['costrecovery']['oil_ftp_is_shared'],
        oil_ftp_portion=convert_to_float(target=data['costrecovery']['oil_ftp_portion']),
        gas_ftp_is_available=data['costrecovery']['gas_ftp_is_available'],
        gas_ftp_is_shared=data['costrecovery']['gas_ftp_is_shared'],
        gas_ftp_portion=convert_to_float(target=data['costrecovery']['gas_ftp_portion']),
        tax_split_type=convert_str_to_taxsplit(str_object=data['costrecovery']['tax_split_type']),
        condition_dict=data['costrecovery']['condition_dict'],
        indicator_rc_icp_sliding=convert_list_to_array_float(
            data_list=data['costrecovery']['indicator_rc_icp_sliding']),
        oil_ctr_pretax_share=convert_to_float(target=data['costrecovery']['oil_ctr_pretax_share']),
        gas_ctr_pretax_share=convert_to_float(target=data['costrecovery']['gas_ctr_pretax_share']),
        oil_ic_rate=convert_to_float(target=data['costrecovery']['oil_ic_rate']),
        gas_ic_rate=convert_to_float(target=data['costrecovery']['gas_ic_rate']),
        ic_is_available=data['costrecovery']['ic_is_available'],
        oil_cr_cap_rate=convert_to_float(target=data['costrecovery']['oil_cr_cap_rate']),
        gas_cr_cap_rate=convert_to_float(target=data['costrecovery']['gas_cr_cap_rate']),
        oil_dmo_volume_portion=convert_to_float(target=data['costrecovery']['oil_dmo_volume_portion']),
        oil_dmo_fee_portion=convert_to_float(target=data['costrecovery']['oil_dmo_fee_portion']),
        oil_dmo_holiday_duration=data['costrecovery']['oil_dmo_holiday_duration'],
        gas_dmo_volume_portion=convert_to_float(target=data['costrecovery']['gas_dmo_volume_portion']),
        gas_dmo_fee_portion=convert_to_float(target=data['costrecovery']['gas_dmo_fee_portion']),
        gas_dmo_holiday_duration=data['costrecovery']['gas_dmo_holiday_duration'],
    )

    # Filling the arguments of the contract with the data input
    contract_arguments_dict = {
        "sulfur_revenue": convert_str_to_otherrevenue(str_object=data['contract_arguments']['sulfur_revenue']),
        "electricity_revenue": convert_str_to_otherrevenue(
            str_object=data['contract_arguments']['electricity_revenue']),
        "co2_revenue": convert_str_to_otherrevenue(str_object=data['contract_arguments']['co2_revenue']),
        "is_dmo_end_weighted": data['contract_arguments']['is_dmo_end_weighted'],
        "tax_regime": convert_str_to_taxregime(str_object=data['contract_arguments']['tax_regime']),
        "effective_tax_rate": convert_list_to_array_float_or_array_or_none(data_list=data['contract_arguments']['effective_tax_rate']),
        "ftp_tax_regime": convert_str_to_ftptaxregime(str_object=data['contract_arguments']['ftp_tax_regime']),
        "sunk_cost_reference_year": data['contract_arguments']['sunk_cost_reference_year'],
        "depr_method": convert_str_to_depremethod(str_object=data['contract_arguments']['depr_method']),
        "decline_factor": data['contract_arguments']['decline_factor'],
        "vat_rate": convert_list_to_array_float_or_array(data_input=data['contract_arguments']['vat_rate']),
        "inflation_rate": convert_list_to_array_float_or_array(data_input=data['contract_arguments']['inflation_rate']),
        "inflation_rate_applied_to": convert_str_to_inflationappliedto(str_object=data['contract_arguments']['inflation_rate_applied_to']),
        "post_uu_22_year2001": True if 'post_uu_22_year2001' not in data['contract_arguments'] else
        data['contract_arguments']['post_uu_22_year2001'],
        "oil_cost_of_sales_applied": False if "oil_cost_of_sales_applied" not in data["contract_arguments"] else
        data["contract_arguments"]["oil_cost_of_sales_applied"],
        "gas_cost_of_sales_applied": False if "gas_cost_of_sales_applied" not in data["contract_arguments"] else
        data["contract_arguments"]["gas_cost_of_sales_applied"],
        "sum_undepreciated_cost": False if 'sum_undepreciated_cost' not in data['contract_arguments'] else
        data['contract_arguments']['sum_undepreciated_cost'],
    }

    # Running the contract
    contract.run(**contract_arguments_dict)

    # Condition when summary is needed
    if summary_result is True:
        # Filling the summary arguments
        summary_arguments_dict = get_summary_dict(data=data)
        summary_arguments_dict['contract'] = contract
        summary = get_summary(**summary_arguments_dict)

        # Converting the summary format to skk summary format
        summary_skk = convert_summary_to_dict(dict_object=summary)

        # Adding the execution info
        summary_skk = add_execution_info(data=summary_skk)

    # Since the required object is only the contract, it will return None for the summary
    else:
        summary_skk = None
        summary_arguments_dict = None

    return summary_skk, contract, contract_arguments_dict, summary_arguments_dict


def get_contract_table(data: dict, contract_type: str = 'Cost Recovery') -> dict:
    """
    Function to get the cash flow table of the contract that has been run.

    Parameters
    ----------
    data: dict
        The dictionary of the data input.
    contract_type: str
        The option for the contract type. The available option are: ['Cost Recovery', 'Gross Split']

    Returns
    -------
    table_all_dict: dict
        The dictionary containing the oil, gas and consolidated cash flow table.
    """
    # Adjusting the variable to the corresponding contract type
    if contract_type == 'Cost Recovery':
        contract = get_costrecovery(data=data)[1]
        year_column = 'Year'

    elif contract_type == 'Gross Split':
        contract = get_grosssplit(data=data)[1]
        year_column = 'Years'

    elif contract_type == 'Base Project':
        contract = get_baseproject(data=data)[1]
        year_column = 'Years'

    else:
        contract = get_transition(data=data)[1]
        year_column = 'Year'

    # Condition when the contract is Transition
    if contract_type == 'Transition':
        # Retrieving the table
        table_oil, table_gas, table_consolidated = get_table(contract=contract)

        # Forming the table dictionary as the output
        table_all_dict = {
            'contract_1': {
                'oil': table_oil[0].set_index(table_oil[0].columns[0]).to_dict(),
                'gas': table_gas[0].set_index(table_gas[0].columns[0]).to_dict(),
                'consolidated': table_consolidated[0].set_index(table_consolidated[0].columns[0]).to_dict()
            },

            'contract_2': {
                'oil': table_oil[1].set_index(table_oil[1].columns[0]).to_dict(),
                'gas': table_gas[1].set_index(table_gas[1].columns[0]).to_dict(),
                'consolidated': table_consolidated[1].set_index(table_consolidated[1].columns[0]).to_dict()
            }
        }

    else:
        # Retrieving the table
        table_oil, table_gas, table_consolidated = get_table(contract=contract)

        # Forming the table dictionary as the output
        table_all_dict = {'oil': table_oil.set_index(year_column).to_dict(),
                          'gas': table_gas.set_index(year_column).to_dict(),
                          'consolidated': table_consolidated.set_index(year_column).to_dict()}

    # Adding the execution info
    table_all_dict = add_execution_info(data=table_all_dict)

    return table_all_dict


def get_contract_optimization(data: dict, contract_type: str = 'Cost Recovery') -> dict:
    """
    The function to run contract optimization. Resulting optimization result in dictionary format.

    Parameters
    ----------
    data: dict
        The dictionary of the data input.
    contract_type: str
        The option for the contract type. The available option are: ['Cost Recovery', 'Gross Split']

    Returns
    -------
    result_parameters: dict
        The result of the optimization in dictionary format

    """
    if 'optimization_arguments' not in data:
        raise ContractException("The payload does not have the optimization_arguments key")

    if data['optimization_arguments'] is None:
        raise ContractException("The payload optimization_arguments does not have any values")

    # Converting the parameters in dict_optimization to the corresponding enum
    optimization_parameters = [
        convert_str_to_optimization_parameters(str_object=i)
        for i in data['optimization_arguments']['dict_optimization']['parameter']
    ]

    # Generating the dictionary of the optimization arguments
    dict_optimization = {
        'parameter': optimization_parameters,
        'min': convert_list_to_array_float(data_list=data['optimization_arguments']['dict_optimization']['min']),
        'max': convert_list_to_array_float(data_list=data['optimization_arguments']['dict_optimization']['max']),
    }

    # Filling the optimization arguments with target_optimization and target_parameter,
    target_optimization_value = data['optimization_arguments']['target_optimization']
    target_parameter = convert_str_to_optimization_targetparameter(
        str_object=data['optimization_arguments']['target_parameter'])

    # Retrieving the contract, contract_arguments_dict, summary_arguments_dict based on the contract type
    if contract_type == 'Cost Recovery':
        contract = get_costrecovery(data=data)[1]
        contract_arguments = get_costrecovery(data=data)[2]
        summary_argument = get_costrecovery(data=data)[3]

    elif contract_type == 'Gross Split':
        contract = get_grosssplit(data=data)[1]
        contract_arguments = get_grosssplit(data=data)[2]
        summary_argument = get_grosssplit(data=data)[3]

    elif contract_type == 'Transition':
        contract = get_transition(data=data)[1]
        contract_arguments = get_transition(data=data)[2]
        summary_argument = get_transition(data=data)[3]

    else:
        contract = NotImplemented
        contract_arguments = NotImplemented
        summary_argument = NotImplemented

    if contract_type == 'Transition':
        # Retrieve the original useful life of the capital cost
        useful_life_original = contract.contract2.capital_cost_total.useful_life.tolist()

        list_str, list_params_value, result_optimization, list_executed_contract = optimize_psc_trans(
            dict_optimization=dict_optimization,
            contract=contract,
            contract_arguments=contract_arguments,
            target_optimization_value=target_optimization_value,
            summary_argument=summary_argument,
            target_parameter=target_parameter,
        )

    else:
        # Retrieve the original useful life of the capital cost
        useful_life_original = contract.capital_cost_total.useful_life.tolist()
        list_str, list_params_value, result_optimization, list_executed_contract = optimize_psc(
            dict_optimization=dict_optimization,
            contract=contract,
            contract_arguments=contract_arguments,
            target_optimization_value=target_optimization_value,
            summary_argument=summary_argument,
            target_parameter=target_parameter,
        )

    # Treatment to add the useful life of optimization into the result
    def get_enum_index(enum_list: list, element: any) -> int | None:
        """
        Function to get the index of the OptimizationParameter.DEPRECIATION_ACCELERATION.

        Parameters
        ----------
        enum_list: list
            The source of list.
        element: any
            The corresponding element
        Returns
        -------
        out : int | None
        """
        try:
            return enum_list.index(element)
        except ValueError:
            return None

    # Get the index of the depreciation optimization parameter
    index_depreciation = get_enum_index(
        enum_list=optimization_parameters,
        element=OptimizationParameter.DEPRECIATION_ACCELERATION)

    # Adding condition of the contract type for retrieving the optimized contract
    if contract_type == 'Transition':
        contract_optimized = list_executed_contract[-1].contract2
    else:
        contract_optimized = list_executed_contract[-1]

    # Adding the information of optimized useful life into the list_params_value
    if index_depreciation is not None:
        optimized_capital_cost = {
            "year": contract_optimized.capital_cost_total.expense_year.tolist(),
            "cost_allocation": contract_optimized.capital_cost_total.cost_allocation,
            "cost": contract_optimized.capital_cost_total.cost.tolist(),
            "pis_year": contract_optimized.capital_cost_total.pis_year.tolist(),
            "useful_life_original": useful_life_original,
            "useful_life_optimized": contract_optimized.capital_cost_total.useful_life.tolist(),
            "description": contract_optimized.capital_cost_total.description
        }

        # Adding optimized_capital_cost into the result of the optimization
        list_params_value[index_depreciation] = {
            "depreciation acceleration": list_params_value[index_depreciation],
            "optimized_useful_life": optimized_capital_cost
        }

    # Forming the optimization result into a dictionary object
    optimization_result = {
        'list_str': list_str,
        'list_params_value': list_params_value,
    }

    # Converting the result into dictionary format
    result_parameters = pd.DataFrame(optimization_result).set_index('list_str').to_dict()
    result_parameters['optimization_result'] = result_optimization

    # Adding the execution info
    result_parameters = add_execution_info(data=result_parameters)

    return result_parameters


def get_grosssplit(data: dict, summary_result: bool = True):
    """
    The function to get the Summary, Gross Split object, contract arguments, and summary arguments used.

    Parameters
    ----------
    data: dict
        The dictionary of the data input.
    summary_result: bool
        The condition if the summary result will be generated or not.

    Returns
    -------
    summary_skk: dict
        The executive summary of the contract.
    contract:
        The Gross Split contract object.
    contract_arguments_dict: dict
        The contract arguments used in running the contract calculation.
    summary_arguments_dic: dict
        The summary arguments used in retrieving the executive summary of the contract.

    """
    start_date, end_date, oil_onstream_date, gas_onstream_date, lifting, tangible, intangible, opex, asr, lbt, cost_of_sales = (
        get_setup_dict(data=data))

    contract = GrossSplit(
        start_date=start_date,
        end_date=end_date,
        oil_onstream_date=oil_onstream_date,
        gas_onstream_date=gas_onstream_date,
        lifting=lifting,
        capital_cost=tangible,
        intangible_cost=intangible,
        opex=opex,
        asr_cost=asr,
        lbt_cost=lbt,
        field_status=data['grosssplit']['field_status'],
        field_loc=data['grosssplit']['field_loc'],
        res_depth=data['grosssplit']['res_depth'],
        infra_avail=data['grosssplit']['infra_avail'],
        res_type=data['grosssplit']['res_type'],
        api_oil=data['grosssplit']['api_oil'],
        domestic_use=data['grosssplit']['domestic_use'],
        prod_stage=data['grosssplit']['prod_stage'],
        co2_content=data['grosssplit']['co2_content'],
        h2s_content=data['grosssplit']['h2s_content'],
        base_split_ctr_oil=convert_to_float(target=data['grosssplit']['base_split_ctr_oil']),
        base_split_ctr_gas=convert_to_float(target=data['grosssplit']['base_split_ctr_gas']),
        split_ministry_disc=convert_to_float(target=data['grosssplit']['split_ministry_disc']),
        oil_dmo_volume_portion=convert_to_float(target=data['grosssplit']['oil_dmo_volume_portion']),
        oil_dmo_fee_portion=convert_to_float(target=data['grosssplit']['oil_dmo_fee_portion']),
        oil_dmo_holiday_duration=data['grosssplit']['oil_dmo_holiday_duration'],
        gas_dmo_volume_portion=convert_to_float(target=data['grosssplit']['gas_dmo_volume_portion']),
        gas_dmo_fee_portion=convert_to_float(target=data['grosssplit']['gas_dmo_fee_portion']),
        gas_dmo_holiday_duration=data['grosssplit']['gas_dmo_holiday_duration'],

    )

    # Filling the arguments of the contract with the data input
    contract_arguments_dict = {
        "sulfur_revenue": convert_str_to_otherrevenue(str_object=data['contract_arguments']['sulfur_revenue']),
        "electricity_revenue": convert_str_to_otherrevenue(
            str_object=data['contract_arguments']['electricity_revenue']),
        "co2_revenue": convert_str_to_otherrevenue(str_object=data['contract_arguments']['co2_revenue']),
        "is_dmo_end_weighted": data['contract_arguments']['is_dmo_end_weighted'],
        "tax_regime": convert_str_to_taxregime(str_object=data['contract_arguments']['tax_regime']),
        "effective_tax_rate": convert_list_to_array_float_or_array_or_none(data_list=data['contract_arguments']['effective_tax_rate']),
        "sunk_cost_reference_year": data['contract_arguments']['sunk_cost_reference_year'],
        "depr_method": convert_str_to_depremethod(str_object=data['contract_arguments']['depr_method']),
        "decline_factor": data['contract_arguments']['decline_factor'],
        "vat_rate": convert_list_to_array_float_or_array(data_input=data['contract_arguments']['vat_rate']),
        "inflation_rate": convert_list_to_array_float_or_array(data_input=data['contract_arguments']['inflation_rate']),
        "inflation_rate_applied_to": convert_str_to_inflationappliedto(
            str_object=data['contract_arguments']['inflation_rate_applied_to']),
        "cum_production_split_offset": convert_list_to_array_float_or_array(data_input=data["contract_arguments"]["cum_production_split_offset"]),
        "amortization": data["contract_arguments"]["amortization"],
        "regime": convert_grosssplitregime_to_enum(target=data["contract_arguments"]["regime"]),
        "sum_undepreciated_cost": False if 'sum_undepreciated_cost' not in data['contract_arguments'] else
        data['contract_arguments']['sum_undepreciated_cost'],
    }

    # Running the contract
    contract.run(**contract_arguments_dict)

    # Condition when summary is needed
    if summary_result is True:
        # Filling the summary arguments
        summary_arguments_dict = get_summary_dict(data=data)
        summary_arguments_dict['contract'] = contract
        summary = get_summary(**summary_arguments_dict)

        # Converting the summary format to skk summary format
        summary_skk = convert_summary_to_dict(dict_object=summary)

        # Adding the execution info
        summary_skk = add_execution_info(data=summary_skk)

    else:
        summary_skk = None
        summary_arguments_dict = None

    return summary_skk, contract, contract_arguments_dict, summary_arguments_dict


def get_transition(data: dict):
    """
    The function to get the Summary, Transition object, contract arguments, and summary arguments used.

    Parameters
    ----------
    data: dict
        The dictionary of the data input.

    Returns
    -------
    summary_skk: dict
        The executive summary of the contract.
    contract:
        The Transition contract object.
    contract_arguments_dict: dict
        The contract arguments used in running the contract calculation.
    summary_arguments_dic: dict
        The summary arguments used in retrieving the executive summary of the contract.

    """
    # Defining contract_1
    if data['contract_1']['costrecovery'] is not None and data['contract_1']['grosssplit'] is None:
        _, contract_1, contract_arguments_1, _ = get_costrecovery(data=data['contract_1'], summary_result=False)

    elif data['contract_1']['grosssplit'] is not None and data['contract_1']['costrecovery'] is None:
        _, contract_1, contract_arguments_1, _ = get_grosssplit(data=data['contract_1'], summary_result=False)

    else:
        raise ContractException("Contract type is not recognized")

    # Defining contract_2
    if data['contract_2']['costrecovery'] is not None and data['contract_2']['grosssplit'] is None:
        _, contract_2, contract_arguments_2, _ = get_costrecovery(data=data['contract_2'], summary_result=False)

    elif data['contract_2']['grosssplit'] is not None and data['contract_2']['costrecovery'] is None:
        _, contract_2, contract_arguments_2, _ = get_grosssplit(data=data['contract_2'], summary_result=False)

    else:
        raise ContractException("Contract type is not recognized")

    # generating the transition contract object
    contract = Transition(contract1=contract_1,
                          contract2=contract_2,
                          argument_contract1=contract_arguments_1,
                          argument_contract2=contract_arguments_2, )

    # Generating the transition contract arguments
    contract_arguments_dict = data['contract_arguments']

    # Running the transition contract
    contract.run(**contract_arguments_dict)

    # Filling the summary arguments
    summary_arguments_dict = get_summary_dict(data=data)
    summary_arguments_dict['contract'] = contract

    summary = get_summary(**summary_arguments_dict)

    # Converting the summary format to skk summary format
    summary_skk = convert_summary_to_dict(dict_object=summary)

    # Adding the execution info
    summary_skk = add_execution_info(data=summary_skk)

    return summary_skk, contract, contract_arguments_dict, summary_arguments_dict


def add_execution_info(data: dict) -> dict:
    """
    Function to adding the execution info into a dictionary.

    Parameters
    ----------
    data: dict
        The dictionary which will added with execution info

    Returns
    -------
    out: dict
        Dictionary containing the execution info
    """
    # Defining the execution date
    execution_date = datetime.now().strftime("%d/%m/%Y %H:%M:%S")

    # Defining the PySCnomics Version
    package_version = " "
    try:
        from importlib.metadata import version
        package_version = version('pyscnomics')

    except:
        pass

    # Parsing the data into the data output
    execution_info = {'execution_datetime': execution_date,
                      'package_version': package_version, }

    data['execution_info'] = execution_info
    return data


def get_detailed_summary(data: dict, contract_type: str):

    if contract_type == 'Cost Recovery':
        summary_args = get_costrecovery(data=data, summary_result=True)[3]

    elif contract_type == 'Gross Split':
        summary_args = get_grosssplit(data=data, summary_result=True)[3]

    elif contract_type == 'Transition':
        summary_args = get_transition(data=data)[3]

    elif contract_type == 'Base Project':
        summary_args = get_baseproject(data=data, summary_result=True)[3]

    else:
        summary_args = None

    return get_summary(**summary_args)


def get_baseproject(data: dict, summary_result: bool = True):
    """
    The function to get the Summary, Base Project object, contract arguments, and summary arguments used.

    Parameters
    ----------
    data: dict
        The dictionary of the data input.
    summary_result: bool
        The condition if the summary result will be generated or not.

    Returns
    -------
    summary_skk: dict
        The executive summary of the contract.
    contract:
        The Base Project contract object.
    contract_arguments_dict: dict
        The contract arguments used in running the contract calculation.
    summary_arguments_dic: dict
        The summary arguments used in retrieving the executive summary of the contract.

    """
    start_date, end_date, oil_onstream_date, gas_onstream_date, lifting, tangible, intangible, opex, asr, lbt, cost_of_sales = (
        get_setup_dict(data=data))

    contract = BaseProject(start_date=start_date,
                           end_date=end_date,
                           oil_onstream_date=oil_onstream_date,
                           gas_onstream_date=gas_onstream_date,
                           lifting=lifting,
                           capital_cost=tangible,
                           intangible_cost=intangible,
                           opex=opex,
                           asr_cost=asr,
                           lbt_cost=lbt)

    contract_arguments_dict = {
        "sulfur_revenue": convert_str_to_otherrevenue(str_object=data['contract_arguments']['sulfur_revenue']),
        "electricity_revenue": convert_str_to_otherrevenue(
            str_object=data['contract_arguments']['electricity_revenue']),
        "co2_revenue": convert_str_to_otherrevenue(str_object=data['contract_arguments']['co2_revenue']),
        "sunk_cost_reference_year": data['contract_arguments']['sunk_cost_reference_year'],
        "tax_rate": convert_list_to_array_float_or_array(data_input=data['contract_arguments']['vat_rate']),
        "inflation_rate": convert_list_to_array_float_or_array(data_input=data['contract_arguments']['inflation_rate']),
        "inflation_rate_applied_to": convert_str_to_inflationappliedto(str_object=data['contract_arguments']['inflation_rate_applied_to']),
    }

    contract.run(**contract_arguments_dict)

    # Condition when summary is needed
    if summary_result is True:
        # Filling the summary arguments
        summary_arguments_dict = get_summary_dict(data=data)
        summary_arguments_dict['contract'] = contract
        summary = get_summary(**summary_arguments_dict)

        # Converting the summary format to skk summary format
        summary_skk = convert_summary_to_dict(dict_object=summary)

        # Adding the execution info
        summary_skk = add_execution_info(data=summary_skk)

    # Since the required object is only the contract, it will return None for the summary
    else:
        summary_skk = None
        summary_arguments_dict = None

    return summary_skk, contract, contract_arguments_dict, summary_arguments_dict


def get_ltp(
        volume: float,
        start_year: int,
        end_year: int,
        fluid_type: FluidType
) -> np.ndarray:
    """
    Function to get the ltp array.

    Parameters
    ----------
    volume:float
        The volume of the reserves.
    start_year:int
        The start year.
    end_year:int
        The end year.
    fluid_type: FluidType
        The FluidType of the corresponding volume.

    Returns
    -------
    out: np.ndarray
        The array of ltp

    """
    # Condition checking for the fluid type for initiating the array of ltp
    if fluid_type == FluidType.OIL:
        rate_ltp = oil_ltp_predict(volume=volume)
    elif fluid_type == FluidType.GAS:
        rate_ltp = gas_ltp_predict(volume=volume)
    else:
        raise LTPModelException(
            f"Unsupported Fluid Type {fluid_type} "
        )

    # Initiating the array of years
    year_arr = np.arange(start_year, end_year + 1)
    gap = len(year_arr) - len(rate_ltp)

    # Checking condition for the given years gap
    if gap < 0:
        raise LTPModelException(
            f"Forecast years from {start_year} to {end_year} is too short. "
            f"Please set the end_year at least until {end_year + (-1 * gap)}"
        )

    rate_gap = np.zeros(gap)
    rate_adjusted = np.concatenate([rate_gap, rate_ltp])

    return rate_adjusted


def get_ltp_dict(data: dict):
    """
    The function to get the list of LTP from the given reserves volume.

    Parameters
    ----------
    data: dict
        The dictionary of the data input.

    Returns
    -------
    """
    volume = data['volume']
    start_year = data['start_year']
    end_year = data['end_year']
    fluid_type = read_fluid_type(fluid=data['fluid_type'])

    ltp_array = get_ltp(
        volume=volume,
        start_year=start_year,
        end_year=end_year,
        fluid_type=fluid_type,)

    return pd.DataFrame(
        {
            'year': np.arange(start_year, end_year + 1).tolist(),
            'ltp': ltp_array.tolist()
        }
    ).set_index('year').to_dict()


def get_rdp(
        start_year: int,
        end_year: int,
        year_rampup: int,
        drate: float,
        q_plateau_ratio: float,
        q_min_ratio: float,
        volume: float
) -> np.ndarray:
    """
    The function to get the RDP array.

    Parameters
    ----------
    start_year
    end_year
    year_rampup
    drate
    q_plateau_ratio
    q_min_ratio
    volume

    Returns
    -------
    out: np.ndarray
        The array of rdp
    """

    # Initiating Model
    model = RPDModel(
        year_rampup=year_rampup,
        drate=drate,
        q_plateau_ratio=q_plateau_ratio,
        q_min_ratio=q_min_ratio,
    )

    rate_rdp = model.predict(volume)

    # Initiating the array of years
    year_arr = np.arange(start_year, end_year + 1)
    gap = len(year_arr) - len(rate_rdp)

    # Checking condition for the given years gap
    if gap < 0:
        raise RDPModelException(
            f"Forecast years from {start_year} to {end_year} is too short. "
            f"Please set the end_year at least until {end_year + (-1 * gap)}"
        )

    rate_gap = np.zeros(gap)
    rate_adjusted = np.concatenate([rate_gap, rate_rdp])

    return rate_adjusted


def get_rpd_dict(data: dict):
    """
    The function to get the list of RPD from the given reserves volume.

    Parameters
    ----------
    data: dict
        The dictionary of the data input.

    Returns
    -------
    """
    # Initiating the data input
    year_rampup = data['year_rampup']
    drate = data['drate']
    q_plateau_ratio = data['q_plateau_ratio']
    q_min_ratio = data['q_min_ratio']
    volume = data['volume']
    start_year = data['start_year']
    end_year = data['end_year']

    # Get the array of rdp
    rdp_array = get_rdp(
        start_year=start_year,
        end_year=end_year,
        year_rampup=year_rampup,
        drate=drate,
        q_plateau_ratio=q_plateau_ratio,
        q_min_ratio=q_min_ratio,
        volume=volume,
    )

    return pd.DataFrame(
        {
            'year': np.arange(start_year, end_year + 1).tolist(),
            'rpd': rdp_array.tolist()
        }
    ).set_index('year').to_dict()

def get_grosssplit_split(data: dict):
    """
    The function to get the contractor split information from Gross Split Contract Scheme.

    Parameters
    ----------
    data: dict
        The dictionary of the data input.

    Returns
    -------
    dict
        The dictionary containing the information of the contractor split.


    """
    start_date, end_date, oil_onstream_date, gas_onstream_date, lifting, tangible, intangible, opex, asr, lbt, cost_of_sales = (
        get_setup_dict(data=data))

    contract = GrossSplit(
        start_date=start_date,
        end_date=end_date,
        oil_onstream_date=oil_onstream_date,
        gas_onstream_date=gas_onstream_date,
        lifting=lifting,
        capital_cost=tangible,
        intangible_cost=intangible,
        opex=opex,
        asr_cost=asr,
        lbt_cost=lbt,
        field_status=data['grosssplit']['field_status'],
        field_loc=data['grosssplit']['field_loc'],
        res_depth=data['grosssplit']['res_depth'],
        infra_avail=data['grosssplit']['infra_avail'],
        res_type=data['grosssplit']['res_type'],
        api_oil=data['grosssplit']['api_oil'],
        domestic_use=data['grosssplit']['domestic_use'],
        prod_stage=data['grosssplit']['prod_stage'],
        co2_content=data['grosssplit']['co2_content'],
        h2s_content=data['grosssplit']['h2s_content'],
        base_split_ctr_oil=convert_to_float(target=data['grosssplit']['base_split_ctr_oil']),
        base_split_ctr_gas=convert_to_float(target=data['grosssplit']['base_split_ctr_gas']),
        split_ministry_disc=convert_to_float(target=data['grosssplit']['split_ministry_disc']),
        oil_dmo_volume_portion=convert_to_float(target=data['grosssplit']['oil_dmo_volume_portion']),
        oil_dmo_fee_portion=convert_to_float(target=data['grosssplit']['oil_dmo_fee_portion']),
        oil_dmo_holiday_duration=data['grosssplit']['oil_dmo_holiday_duration'],
        gas_dmo_volume_portion=convert_to_float(target=data['grosssplit']['gas_dmo_volume_portion']),
        gas_dmo_fee_portion=convert_to_float(target=data['grosssplit']['gas_dmo_fee_portion']),
        gas_dmo_holiday_duration=data['grosssplit']['gas_dmo_holiday_duration'],

    )

    # Filling the arguments of the contract with the data input
    contract_arguments_dict = {
        "sulfur_revenue": convert_str_to_otherrevenue(str_object=data['contract_arguments']['sulfur_revenue']),
        "electricity_revenue": convert_str_to_otherrevenue(
            str_object=data['contract_arguments']['electricity_revenue']),
        "co2_revenue": convert_str_to_otherrevenue(str_object=data['contract_arguments']['co2_revenue']),
        "is_dmo_end_weighted": data['contract_arguments']['is_dmo_end_weighted'],
        "tax_regime": convert_str_to_taxregime(str_object=data['contract_arguments']['tax_regime']),
        "effective_tax_rate": convert_list_to_array_float_or_array_or_none(data_list=data['contract_arguments']['effective_tax_rate']),
        "sunk_cost_reference_year": data['contract_arguments']['sunk_cost_reference_year'],
        "depr_method": convert_str_to_depremethod(str_object=data['contract_arguments']['depr_method']),
        "decline_factor": data['contract_arguments']['decline_factor'],
        "vat_rate": convert_list_to_array_float_or_array(data_input=data['contract_arguments']['vat_rate']),
        "inflation_rate": convert_list_to_array_float_or_array(data_input=data['contract_arguments']['inflation_rate']),
        "inflation_rate_applied_to": convert_str_to_inflationappliedto(
            str_object=data['contract_arguments']['inflation_rate_applied_to']),
        "cum_production_split_offset": convert_list_to_array_float_or_array(data_input=data["contract_arguments"]["cum_production_split_offset"]),
        "amortization": data["contract_arguments"]["amortization"],
        "regime": convert_grosssplitregime_to_enum(target=data["contract_arguments"]["regime"])
    }

    # Running the contract
    contract.run(**contract_arguments_dict)

    # Retrieving the split information
    contractor_split = pd.DataFrame({
        'project_years': contract.project_years.tolist(),
        'oil_base_split': contract._oil_base_split.tolist(),
        'gas_base_split': contract._gas_base_split.tolist(),
        'var_split_array': contract._var_split_array.tolist(),
        'oil_prog_price_split': contract._oil_prog_price_split.tolist(),
        'gas_prog_price_split': contract._gas_prog_price_split.tolist(),
        'oil_prog_cumulative_production_split': contract._oil_prog_cum_split.tolist(),
        'gas_prog_cumulative_production_split': contract._gas_prog_cum_split.tolist(),
        'oil_prog_total_split': contract._oil_prog_split.tolist(),
        'gas_prog_total_split': contract._gas_prog_split.tolist(),
        'oil_ctr_split': contract._oil_ctr_split_prior_bracket.tolist(),
        'gas_ctr_split': contract._gas_ctr_split_prior_bracket.tolist(),
    }).set_index('project_years').to_dict()

    years_of_maximum_split = {
        'oil': contract._oil_year_maximum_ctr_split.tolist(),
        'gas': contract._gas_year_maximum_ctr_split.tolist(),
    }

    return {
        'contractor_split': contractor_split,
        'years_of_maximum_split': years_of_maximum_split
    }


def get_transition_split(data: dict):
    """
    The function to get the contractor split information from Transition Contract Scheme.

    Parameters
    ----------
    data: dict
        The dictionary of the data input.

    Returns
    -------
    dict
        The dictionary containing the information of the contractor split.

    """
    # Defining contract_1
    if data['contract_1']['costrecovery'] is not None and data['contract_1']['grosssplit'] is None:
        _, contract_1, contract_arguments_1, _ = get_costrecovery(data=data['contract_1'], summary_result=False)

    elif data['contract_1']['grosssplit'] is not None and data['contract_1']['costrecovery'] is None:
        _, contract_1, contract_arguments_1, _ = get_grosssplit(data=data['contract_1'], summary_result=False)

    else:
        raise ContractException("Contract type is not recognized")

    # Defining contract_2
    if data['contract_2']['costrecovery'] is not None and data['contract_2']['grosssplit'] is None:
        _, contract_2, contract_arguments_2, _ = get_costrecovery(data=data['contract_2'], summary_result=False)

    elif data['contract_2']['grosssplit'] is not None and data['contract_2']['costrecovery'] is None:
        _, contract_2, contract_arguments_2, _ = get_grosssplit(data=data['contract_2'], summary_result=False)

    else:
        raise ContractException("Contract type is not recognized")

    # generating the transition contract object
    contract = Transition(contract1=contract_1,
                          contract2=contract_2,
                          argument_contract1=contract_arguments_1,
                          argument_contract2=contract_arguments_2, )

    # Generating the transition contract arguments
    contract_arguments_dict = data['contract_arguments']

    # Running the transition contract
    contract.run(**contract_arguments_dict)

    # Making the base for the loops and container of the result
    result = {}

    # Retrieving the split information from first contract
    for index, contract in enumerate([contract_1, contract_2]):
        if isinstance(contract, GrossSplit):
            # Retrieving the split information
            contractor_split = pd.DataFrame({
                'project_years': contract.project_years.tolist(),
                'oil_base_split': contract._oil_base_split.tolist(),
                'gas_base_split': contract._gas_base_split.tolist(),
                'var_split_array': contract._var_split_array.tolist(),
                'oil_prog_split': contract._oil_prog_split.tolist(),
                'gas_prog_split': contract._gas_prog_split.tolist(),
                'oil_ctr_split': contract._oil_ctr_split_prior_bracket.tolist(),
                'gas_ctr_split': contract._gas_ctr_split_prior_bracket.tolist(),
            }).set_index('project_years').to_dict()

            years_of_maximum_split = {
                'oil': contract._oil_year_maximum_ctr_split.tolist(),
                'gas': contract._gas_year_maximum_ctr_split.tolist(),
            }

            result['contract_' + str(index+1)] = {
                'contractor_split': contractor_split,
                'years_of_maximum_split': years_of_maximum_split
            }
        else:
            result['contract_' + str(index + 1)] = {}
            pass

    return result

def get_economic_limit(
        data: dict,
):
    """
    The function to get the information of economic limit years from a cashflow using selected method.

    Parameters
    ----------
    data: dict

    Returns
    -------
    int
        The index
    """
    years = np.array(data['years'], dtype=int)
    cash_flow = np.array(data['cash_flow'], dtype=float)
    method = convert_to_method_limit(target=data['method'])
    index_limit = econ_limit(cashflow=cash_flow, method=method)
    return years[index_limit]


def get_asr_expenditures(data:dict) -> dict:
    """
    The Function to get the expenditures of an ASR cost.

    Parameters
    ----------
    data: dict

    Returns
    -------
    dict
        The dictionary of ASR expenditures.

    """
    # Initiating the asr data
    asr_pseudo = {'asr':data['asr']}

    # Mimics the baseproject data
    data_pseudo = {
        "setup": {
            "start_date": data['start_date'],
            "end_date": data['end_date'],
            "oil_onstream_date": None,
            "gas_onstream_date": None,
        },
        "summary_arguments":{
            "reference_year":None,
            "inflation_rate":0.0,
            "discount_rate": 0.1,
            "npv_mode": "Full Cycle Nominal Terms",
            "discounting_mode": "Mid Year",
            "profitability_discounted": False,
        },
        "contract_arguments":{
            "sulfur_revenue": "Addition to Gas Revenue",
            "electricity_revenue": "Addition to Oil Revenue",
            "co2_revenue": "Addition to Gas Revenue",
            "sunk_cost_reference_year": None,
            "year_inflation": 0,
            "inflation_rate": 0,
            "vat_rate": 0,
            "inflation_rate_applied_to": "CAPEX",
        },
        "lifting": None,
        "capital": None,
        "intangible": None,
        "opex": None,
        "asr": asr_pseudo,
        "lbt": None,
        "cost_of_sales": None,
    }

    # Parsing the data into base project dataclass
    contract = get_baseproject(data=data_pseudo, summary_result=False)[1]

    # Returning the ASR Expenditures
    df = pd.DataFrame(
        {
            'project_years': contract.project_years,
            'oil_asr_expenditures': contract._oil_asr_expenditures_post_tax,
            'gas_asr_expenditures': contract._gas_asr_expenditures_post_tax,
        }
    )
    df = df.set_index('project_years').to_dict()
    return df


def get_lbt_expenditures(data:dict) -> dict:
    """
    The Function to get the expenditures of an LBT cost.

    Parameters
    ----------
    data: dict

    Returns
    -------
    dict
        The dictionary of LBT expenditures.

    """
    # Initiating the LBT data
    lbt_pseudo = {'lbt':data['lbt']}

    # Mimics the baseproject data
    data_pseudo = {
        "setup": {
            "start_date": data['start_date'],
            "end_date": data['end_date'],
            "oil_onstream_date": None,
            "gas_onstream_date": None,
        },
        "summary_arguments":{
            "reference_year":None,
            "inflation_rate":0.0,
            "discount_rate": 0.1,
            "npv_mode": "Full Cycle Nominal Terms",
            "discounting_mode": "Mid Year",
            "profitability_discounted": False,
        },
        "contract_arguments":{
            "sulfur_revenue": "Addition to Gas Revenue",
            "electricity_revenue": "Addition to Oil Revenue",
            "co2_revenue": "Addition to Gas Revenue",
            "sunk_cost_reference_year": None,
            "year_inflation": 0,
            "inflation_rate": 0,
            "vat_rate": 0,
            "inflation_rate_applied_to": "CAPEX",
        },
        "lifting": None,
        "capital": None,
        "intangible": None,
        "opex": None,
        "asr": None,
        "lbt": lbt_pseudo,
        "cost_of_sales": None,
    }

    # Parsing the data into base project dataclass
    contract = get_baseproject(data=data_pseudo, summary_result=False)[1]

    # Returning the LBT Expenditures
    df = pd.DataFrame(
        {
            'project_years': contract.project_years,
            'oil_lbt_expenditures': contract._oil_lbt_expenditures_post_tax,
            'gas_lbt_expenditures': contract._gas_lbt_expenditures_post_tax,
        }
    )
    df = df.set_index('project_years').to_dict()
    return df




import pandas as pd
import numpy as np
import xlwings as xw

from pyscnomics.contracts.project import BaseProject
from pyscnomics.contracts.costrecovery import CostRecovery
from pyscnomics.contracts.grossplit import GrossSplit
from pyscnomics.contracts.transition import Transition

from pyscnomics.tools.table import get_table


def write_cashflow(workbook_object: xw.Book,
                   sheet_name: str,
                   contract: BaseProject | CostRecovery | GrossSplit | Transition,
                   oil_starting_cell: str = 'B5',
                   gas_starting_cell: str = 'B59',
                   consolidated_starting_cell: str = 'B113'):
    """
    The function to write contract cashflow into Excel Workbook.

    Parameters
    ----------
    workbook_object: xw.Book
        The path of the Excel workbook
    sheet_name: str
        The sheet name of where the cashflow is located within the workbook
    contract: CostRecovery | GrossSplit | Transition
        The object of the contract
    oil_starting_cell: str
        The range cell where the oil table cashflow first row and first column is located
    gas_starting_cell: str
        The range cell where the gas table cashflow first row and first column is located
    consolidated_starting_cell: str
        The range cell where the consolidated table cashflow first row and first column is located

    """
    # Producing the table
    df_oil, df_gas, df_consolidated = get_table(contract=contract)

    # Defining the workbook and its sheet when the contract is Transition
    if sheet_name == 'Transition':
        # Defining contract type for frst and second contract
        if isinstance(contract.contract1, CostRecovery) and isinstance(contract.contract2, CostRecovery):
            sheet_cf_1 = 'Result Table CR'
            sheet_cf_2 = 'Result Table CR (2)'

        elif isinstance(contract.contract1, CostRecovery) and isinstance(contract.contract2, GrossSplit):
            sheet_cf_1 = 'Result Table CR'
            sheet_cf_2 = 'Result Table GS'

        elif isinstance(contract.contract1, GrossSplit) and isinstance(contract.contract2, GrossSplit):
            sheet_cf_1 = 'Result Table GS'
            sheet_cf_2 = 'Result Table GS (2)'

        else:
            sheet_cf_1 = 'Result Table GS'
            sheet_cf_2 = 'Result Table CR'

        # Looping the sheet name to write the cashflow into Excel file
        lst_of_sheet = [sheet_cf_1, sheet_cf_2]
        for index, sheet in enumerate(lst_of_sheet):
            # Defining the workbook and its sheet
            ws = workbook_object.sheets(sheet)

            # Writing oil df_oil
            ws.range(oil_starting_cell).value = df_oil[index].values

            # Writing oil df_gas
            ws.range(gas_starting_cell).value = df_gas[index].values

            # Writing oil df_consolidated
            ws.range(consolidated_starting_cell).value = df_consolidated[index].values

    # When the contract is Cost Recovery or Gross Split
    else:
        # Defining the workbook and its sheet
        ws = workbook_object.sheets(sheet_name)

        # Writing oil df_oil
        ws.range(oil_starting_cell).value = df_oil.values

        # Writing oil df_gas
        ws.range(gas_starting_cell).value = df_gas.values

        # Writing oil df_consolidated
        ws.range(consolidated_starting_cell).value = df_consolidated.values


def write_summary(summary_dict: dict,
                  workbook_object: xw.Book,
                  sheet_name: str = 'Summary',
                  range_cell: str = 'E5'):
    """
    A function to generate the summary and write it on the Microsoft Excel sheet.

    Parameters
    ----------
    summary_dict :  dict
        The dictionary containing the argument for generating the summary.
    workbook_object : xw.Book
        The directory path of the Microsoft Excel file.
    sheet_name : str
        The name of the worksheet
    range_cell : str
        The cell range of the summary in the Excel sheet

    Returns
    -------

    """

    # Defining the result that used in summary table
    result_list = [
        summary_dict['lifting_oil'],
        summary_dict['oil_wap'],
        summary_dict['lifting_gas'],
        summary_dict['gas_wap'],
        summary_dict['gross_revenue'],
        None,
        summary_dict['ctr_gross_share'],
        summary_dict['gov_gross_share'],
        summary_dict['sunk_cost'],
        summary_dict['investment'],
        summary_dict['tangible'],
        summary_dict['intangible'],
        summary_dict['opex_and_asr'],
        summary_dict['opex'],
        summary_dict['asr'],
        summary_dict['cost_recovery / deductible_cost'],
        summary_dict['cost_recovery_over_gross_rev'],
        summary_dict['unrec_cost'],
        summary_dict['unrec_over_gross_rev'],
        None,
        summary_dict['ctr_net_share'],
        summary_dict['ctr_net_share_over_gross_share'],
        summary_dict['ctr_net_cashflow'],
        summary_dict['ctr_net_cashflow_over_gross_rev'],
        summary_dict['ctr_npv'],
        summary_dict['ctr_irr'],
        summary_dict['ctr_pot'],
        summary_dict['ctr_pv_ratio'],
        summary_dict['ctr_pi'],
        None,
        summary_dict['gov_gross_share'],
        summary_dict['gov_ftp_share'],
        summary_dict['gov_ddmo'],
        summary_dict['gov_tax_income'],
        summary_dict['gov_take'],
        summary_dict['gov_take_over_gross_rev'],
        summary_dict['gov_take_npv']
    ]

    # Initiating the dataframe that containing the summary
    df_summary = pd.DataFrame(result_list, columns=['control'], dtype=str)

    # Replacing NAN with --- to fill the gap between Excel rows
    df_summary.fillna('---', inplace=True)

    # Selecting a sheet in Excel
    ws = workbook_object.sheets(sheet_name)

    # Test Write
    ws.range(range_cell).value = df_summary.values


def write_opt(list_str: list,
              list_params_value: list,
              result_optimization: float,
              workbook_object: xw.Book,
              sheet_name: str = 'Optimization',
              range_opt_result: str = 'P2',
              range_list_params: str = 'N5',
              range_list_value: str = 'P5',
              ):
    """
    The function to write optimization result into Excel workbook.

    Parameters
    ----------
    list_str: list
        The list of optimization parameter
    list_params_value: list
        The list of optimized parameter value
    result_optimization: float
        The value result of optimized target
    workbook_object: xw.Book
        The workbook object of the Excel file
    sheet_name: str
        The sheet name where the optimization result will be printed on
    range_opt_result: str
        The range cell where the optimization result will be printed on
    range_list_params: str
        The range cell where the optimized parameters will be printed on
    range_list_value: str
        The range cell where the optimized parameters value will be printed on

    """
    # Making dataframe of the optimization parameter list, optimization parameter values
    # df_optim = pd.DataFrame()
    # df_optim['list_parameter'] = list_str
    # df_optim['list_parameter_value'] = list_params_value

    # Defining the workbook and its sheet
    ws = workbook_object.sheets(sheet_name)

    # Writing result of optimization
    ws.range(range_opt_result).value = result_optimization

    # Deleting the row of the optimization result
    ws.range(range_list_params).options(transpose=True).value = [''] * 10
    ws.range(range_list_value).options(transpose=True).value = [''] * 10

    # Writing list of parameter optimization and parameter value of optimization
    ws.range(range_list_params).options(transpose=True).value = list_str
    ws.range(range_list_value).options(transpose=True).value = list_params_value


def write_table(workbook_object: xw.Book,
                sheet_name: str,
                starting_cell: str,
                table: pd.DataFrame | list[pd.DataFrame]):
    """
    The function to write a table into Excel file
    Parameters
    ----------
    workbook_object: xw.Book
        The workbook object of the Excel file
    sheet_name: str
        The sheet name where the table will be printed on
    starting_cell: str
        The starting cell range of where the table will be printed on
    table: pd.Dataframe
        The dataframe object of the table

    """
    # Defining the workbook and its sheet
    ws = workbook_object.sheets(sheet_name)

    # Writing the dataframe into workbook
    ws.range(starting_cell).value = table.values







